"""
This example shows how to train a MEGAN model based on a chemistry dataset consisting of a CSV files
containing the SMILES representations of the molecules.

"""
import os
import warnings
warnings.filterwarnings("ignore")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import csv
import random
from typing import List, Dict, Any

from pycomex.experiment import Experiment
from pycomex.util import Skippable

import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
import tensorflow.keras as ks
from matplotlib.backends.backend_pdf import PdfPages
from kgcnn.data.moleculenet import OneHotEncoder
from kgcnn.data.utils import ragged_tensor_from_nested_numpy
from imageio.v2 import imread
from sklearn.metrics import r2_score

import graph_attention_student.typing as tc
from graph_attention_student.util import DATASETS_FOLDER
from graph_attention_student.data import create_molecule_eye_tracking_dataset
from graph_attention_student.data import load_eye_tracking_dataset
from graph_attention_student.data import load_eye_tracking_dataset_dict
from graph_attention_student.data import process_graph_dataset
from graph_attention_student.training import NoLoss
from graph_attention_student.training import LogProgressCallback
from graph_attention_student.models import MultiAttentionStudent
from graph_attention_student.models import Megan
from graph_attention_student.visualization import plot_node_importances
from graph_attention_student.visualization import plot_edge_importances

# == DATASET PARAMETERS ==
# [!] Insert the absolute path to the CSV file here
CSV_PATH = os.path.join(DATASETS_FOLDER, 'dataset-E.csv')
# CSV_PATH = os.path.join(DATASETS_FOLDER, 'Water', 'Water.csv')
CSV_PATH = os.path.join(DATASETS_FOLDER, 'Benzene', 'Benzene.csv')

# [!] Insert the column name of the smiles column here
SMILES_COLUMN = "SMILES"
# [!] Insert the column name of the target value here
TARGET_COLUMN = 'Solubility'
TRAIN_SPLIT = 0.80

# == MODEL PARAMETERS ==

# This list defines how many graph convolutional layers to configure the network with. Every element adds
# one layer. The numbers themselves are the hidden units to be used for those layers.
UNITS = [3, 3, 3]

# The dropout rate which is applied after EVERY graph convolutional layer of the network. Especially for
# large networks (>20 hidden units and multiple importance channels, this dropout proves very useful)
DROPOUT_RATE = 0.0

# This is the weight of the additional explanation-only step which is being applied to the network.
# This explanation only step is important to develop actually interpretable explanations. Refer to the
# paper for more details about this.
IMPORTANCE_FACTOR = 1.0

# This is another hyperparameter of the explanation only train step. Usually should be between 1 and 10
IMPORTANCE_MULTIPLIER = 10

# This is the number of explanation channels that are generated by the model. This is also the number of
# attention heads used in the graph convolutional layers of the network. So to get a "wider" network, this
# parameter can also be increased. However, note that the value can only be != 2 if importance factor is
# set to exactly 0.0!
IMPORTANCE_CHANNELS = 2

# The coefficient value of the explanation sparsity regularization that is applied to the network. Higher
# values should lead to sparser explanations.
SPARSITY_FACTOR = 1.0

# We need to supply the range of possible target values and a reference target value a priori to the
# network for the regression case. The regression limits should be as complete as possible. The reference
# does not have to be in the middle, changing it will influence the development of the explanations quite
# a bit.
REGRESSION_LIMITS = [-6, +2]
REGRESSION_REFERENCE = 0

# At the tail end of the network there is a MLP, which does the final prediction. This list defines the
# depth of that MLP. Each element adds a layer and the values themselves dictate the hidden values.
# In the regression case a final entry with "1" hidden layer as the output layer will implicitly be
# added.
FINAL_UNITS = [3, 1]

# This dropout is applied after EVERY layer in the final MLP. Using this should not be necessary
FINAL_DROPOUT = 0.0

# == TRAINING PARAMETERS ==
DEVICE = 'gpu:0'
LOSS = ks.losses.MeanSquaredError()
LEARNING_RATE = 0.001
# LEARNING_RATE = 0.01
BATCH_SIZE = 32
EPOCHS = 500

# == EVALUATION PARAMETERS ==
LOG_STEP = 20
METRIC_KEY = 'mean_squared_error'
CHANNEL_NAMES = {
    0: 'Negative',
    1: 'Positive'
}
CHANNEL_DIRECTIONS = {
    0: -1,
    1: 1,
}

DEBUG = True
BASE_PATH = os.getcwd()
NAMESPACE = 'results/smiles_example'
with Skippable(), (e := Experiment(base_path=BASE_PATH, namespace=NAMESPACE, glob=globals())):

    # -- CONVERTING TO "EYE TRACKING DATASET"
    # First of all we convert the dataset into the "eye tracking" format. In this format, the dataset
    # is represented as a folder where each sample is associated with 2 files:
    # - One PNG file which visualizes that element
    # - One JSON file, which contains the graph representation of the molecule and other metadata
    folder_path = os.path.dirname(CSV_PATH)
    dataset_path = os.path.join(folder_path, 'eye_tracking_dataset')

    # We only do the conversion though if it does not already exists
    if not os.path.exists(dataset_path):
        e.info('starting to process dataset...')
        e.info('reading csv file...')
        molecule_infos: Dict[str, Dict[str, Any]] = {}
        with open(CSV_PATH, mode='r') as csv_file:
            reader = csv.DictReader(csv_file, delimiter=',', quotechar='"')
            for i, row in enumerate(reader):
                smile = row[SMILES_COLUMN]
                value = float(row[TARGET_COLUMN])
                element_id = str(i)

                molecule_info = {
                    'id': element_id,
                    'name': element_id,
                    'value': value,
                    'smiles': smile,
                    'input_type': 'regression',
                }
                molecule_infos[element_id] = molecule_info
                if i % LOG_STEP == 0:
                    e.info(f' * ({i:03d}) - smile: {smile}')

        e.info(f'creating eye tracking dataset in: {dataset_path}')
        os.mkdir(dataset_path)
        create_molecule_eye_tracking_dataset(
            molecule_infos=molecule_infos,
            dest_path=dataset_path,
            logger=e.logger,
            log_step=LOG_STEP,
            # This dictionary determines what kinds of chemical properties are going to be added as features
            # of the resulting dataset. There are two kinds of features: Ones associated with the
            # nodes (atoms) of the graph and the ones associated with the edges (bonds).
            # This dictionary will be passed almost exactly as it is to the "kgcnn.moleculenet.MoleculeNet"
            # class which does the actual interaction with RDKit, so look into that for more details about
            # what kinds of options are available.
            set_attributes_kwargs={
                'nodes': ['Symbol', 'TotalDegree', 'FormalCharge', 'NumRadicalElectrons', 'Hybridization',
                          'IsAromatic', 'IsInRing', 'TotalNumHs', 'CIPCode', 'ChiralityPossible',
                          'ChiralTag'],
                'encoder_nodes': {
                    'Symbol': OneHotEncoder(
                        ['B', 'C', 'N', 'O', 'F', 'Si', 'P', 'S', 'Cl', 'As', 'Se', 'Br', 'Te', 'I', 'At'],
                        dtype="str"
                    ),
                    'Hybridization': OneHotEncoder([2, 3, 4, 5, 6]),
                    'TotalDegree': OneHotEncoder([0, 1, 2, 3, 4, 5], add_unknown=False),
                    'TotalNumHs': OneHotEncoder([0, 1, 2, 3, 4], add_unknown=False),
                    'CIPCode': OneHotEncoder(['R', 'S'], add_unknown=False, dtype='str'),
                    'ChiralityPossible': OneHotEncoder(["1"], add_unknown=False, dtype='str'),
                },
                'edges': ['BondType', 'IsAromatic', 'IsConjugated', 'IsInRing', 'Stereo'],
                'encoder_edges': {
                    'BondType': OneHotEncoder([1, 2, 3, 12], add_unknown=False),
                    'Stereo': OneHotEncoder([0, 1, 2, 3], add_unknown=False)
                },
            }
        )
        e.info(f'created eye tracking dataset with {len(molecule_infos)} elements!')

    # -- LOADING DATASET
    # Now we need to load the dataset and do some pre-processing so that we can use it for model training
    e.info('loading dataset...')
    eye_tracking_dataset: tc.EtDatasetDict = load_eye_tracking_dataset_dict(dataset_path)
    e.info(f'loaded dataset with {len(eye_tracking_dataset)} elements')

    # the "eye_tracking_dataset" consists of dicts, which hold the information about the filepaths for the
    # image file, metadata file, the metadata dict itself, the graph dict. But for the training we only
    # need the graph dict, which is why we extract those here.
    dataset = []
    for c, (name, data) in enumerate(eye_tracking_dataset.items()):
        metadata = data['metadata']
        g = metadata['graph']

        # We need to fill this field with the target value! In the future for multiple regression targets
        # this may also be a numpy array / vector. The important thing is that the output shape of the
        # network matches the shape of this!
        g['graph_labels'] = np.array([data['metadata']['value']])
        metadata['index'] = c

        dataset.append(g)

    dataset_indices = list(range(len(dataset)))
    dataset_length = len(dataset)
    train_indices = random.sample(dataset_indices, k=int(TRAIN_SPLIT * len(dataset_indices)))
    test_indices = [index for index in dataset_indices if index not in train_indices]
    e['train_indices'] = train_indices
    e['test_indices'] = test_indices
    dataset_index_map = {d['metadata']['index']: d for d in eye_tracking_dataset.values()}

    # This turns the list of graph dicts into the final form which we need for the training of the model:
    # keras RaggedTensors which contain all the graphs.
    x_train, y_train, x_test, y_test = process_graph_dataset(dataset, test_indices)

    # -- TRAINING THE MODEL
    model = Megan(
        units=UNITS,
        final_units=FINAL_UNITS,
        use_bias=True,
        importance_factor=IMPORTANCE_FACTOR,
        sparsity_factor=SPARSITY_FACTOR,
        regression_limits=REGRESSION_LIMITS,
        regression_reference=REGRESSION_REFERENCE,
        dropout_rate=DROPOUT_RATE
    )

    model.compile(
        loss=[
            ks.losses.MeanSquaredError(),
            NoLoss(),
            NoLoss()
        ],
        loss_weights=[
            1,
            0,
            0
        ],
        metrics=[
            ks.metrics.MeanSquaredError(),
            ks.metrics.MeanAbsoluteError(),
        ],
        optimizer=ks.optimizers.Adam(learning_rate=LEARNING_RATE),
        run_eagerly=False
    )
    with tf.device('gpu:0'):
        history = model.fit(
            x_train,
            y_train,
            batch_size=BATCH_SIZE,
            epochs=EPOCHS,
            validation_data=(x_test, y_test),
            validation_freq=1,
            callbacks=LogProgressCallback(
                logger=e.logger,
                epoch_step=LOG_STEP,
                identifier=f'val_output_1_{METRIC_KEY}'
            ),
            verbose=0
        )

    e[f'history'] = history.history
    e[f'epochs'] = list(range(EPOCHS))
    param_count = model.count_params()
    e['param_count'] = param_count
    e.info(f'number of parameters: {param_count}')

    # -- EVALUATING TEST SET
    # Now we get to the reason why we converted the dataset into the "eye tracking" format first: This will
    # now simplify the visualization of the results a lot!
    # We now evaluate all the elements in the test set and visualize the results. We will do this by
    # creating one big PDF document which displays all the elements of the test set and then also displays
    # what kinds of explanations the model predicted for those.
    #e.info('Evaluating test set')

    e.info('Evaluating test set and drawing examples...')
    pdf_path = os.path.join(e.path, f'explanations.pdf')
    ncols = 1
    nrows = IMPORTANCE_CHANNELS
    with PdfPages(pdf_path) as pdf:

        e.info('generating test set predictions...')
        out_pred, ni_pred, ei_pred = [v.numpy() for v in model(x_test)]
        out_true, ni_true, ei_true = y_test

        # Saving the values for the current element to the internal experiment dictionary
        e.info('saving predictions...')
        for c, index in enumerate(test_indices):
            e[f'out/true/{index}'] = float(out_true[c][0])
            e[f'out/pred/{index}'] = float(out_pred[c][0])
            ni = ni_pred[c]
            ei = ei_pred[c]
            e[f'ni/{index}'] = ni
            e[f'ei/{index}'] = ei

        # Calculating the explanation fidelity* value
        if IMPORTANCE_FACTOR != 0:
            for k, direction in CHANNEL_DIRECTIONS.items():
                e.info(f'deviation for channel {k} with intended direction: {direction}')
                masks = []
                for ni in ni_pred:
                    mask = np.ones_like(ni)
                    mask[:, k] = 0
                    masks.append(mask)

                masks_tensor = ragged_tensor_from_nested_numpy(masks)
                out_masked, _, _ = model(x_test, node_importances_mask=masks_tensor)
                out_masked = out_masked.numpy()
                for c, index in enumerate(test_indices):
                    e[f'out_mod_fidelity/{index}/{k}'] = float(out_masked[c][0])

            e.info('calculating fidelity* values...')
            for index in test_indices:
                out = e[f'out/pred/{index}']
                fidelity = 0
                for k in CHANNEL_DIRECTIONS.keys():
                    out_ = e[f'out_mod_fidelity/{index}/{k}']
                    fidelity += CHANNEL_DIRECTIONS[k] * (out - out_)
                    e[f'fidelity/{index}'] = fidelity

        # Plotting the explanations for the current sample
        e.info('plotting the examples...')
        for c, index in enumerate(test_indices):
            data = dataset_index_map[index]
            g = data['metadata']['graph']
            image = np.asarray(imread(data['image_path']))
            fig, rows = plt.subplots(nrows=nrows, ncols=ncols, figsize=(8 * ncols, 8 * nrows), squeeze=False)
            fig.suptitle(f'Element {data["metadata"]["name"]}\n'
                         f'{data["metadata"]["smiles"]}\n'
                         f'True: {out_true[c]} - Pred: {out_pred[c]}\n'
                         f'Explanation Fidelity: {e[f"fidelity/{index}"]:.3f}')
            node_coordinates = np.array(g['node_coordinates'])
            for row_index in range(nrows):
                column_index = 0

                ax = rows[row_index][column_index]
                ax.set_title(f'fidelity deviation: {e[f"out_mod_fidelity/{index}/{row_index}"]:.3f}')
                ax.set_ylabel(f'Explanation Channel {row_index}:\n'
                              f'{CHANNEL_NAMES[row_index]}')
                ax.imshow(image, extent=(0, image.shape[0], 0, image.shape[1]))
                plot_node_importances(
                    g=g,
                    ax=ax,
                    vmax=np.max(ni),
                    node_importances=ni[:, row_index],
                    node_coordinates=node_coordinates
                )
                plot_edge_importances(
                    g=g,
                    ax=ax,
                    vmax=np.max(ei),
                    edge_importances=ei[:, row_index],
                    node_coordinates=node_coordinates
                )

            pdf.savefig(fig)
            plt.close(fig)

            if c % LOG_STEP == 0:
                e.info(f' * ({c}/{len(test_indices)})')


with Skippable(), e.analysis:

    # -- PLOTTING TRAINING PROCESS
    e.info('plotting training information...')
    pdf_path = os.path.join(e.path, 'training.pdf')
    with PdfPages(pdf_path) as pdf:
        epochs = list(range(EPOCHS))
        # Plotting the loss of the training and validation dataset
        fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(8, 8))
        fig.suptitle('Loss during Training')
        ax.set_ylabel('Loss')
        ax.set_xlabel('Epochs')
        ax.plot(epochs, e['history']['output_1_loss'],
                color='blue', ls='--', alpha=0.5, label='train')
        final_value = e['history/output_1_loss'][-1]
        ax.scatter(epochs[-1], final_value, color='blue', alpha=0.5, label=f'{final_value:.2f}')
        ax.plot(epochs, e['history']['val_output_1_loss'],
                color='blue', ls='-', alpha=1.0, label='test')
        final_value = e['history/val_output_1_loss'][-1]
        ax.scatter(epochs[-1], final_value, color='blue', alpha=1.0, label=f'{final_value:.2f}')
        ax.legend()
        pdf.savefig(fig)
        plt.close(fig)

        # Plotting the metric of the training and validation dataset
        fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(8, 8))
        fig.suptitle(f'"{METRIC_KEY}" during training')
        ax.set_ylabel(METRIC_KEY)
        ax.set_xlabel('Epochs')
        ax.plot(epochs, e[f'history/output_1_{METRIC_KEY}'],
                color='blue', ls='--', alpha=0.5, label='train')
        final_value = e[f'history/output_1_{METRIC_KEY}'][-1]
        ax.scatter(epochs[-1], final_value, color='blue', alpha=0.5, label=f'{final_value:.2f}')
        ax.plot(epochs, e[f'history/val_output_1_{METRIC_KEY}'],
                color='blue', ls='-', alpha=1.0, label='test')
        final_value = e[f'history/val_output_1_{METRIC_KEY}'][-1]
        ax.scatter(epochs[-1], final_value, color='blue', alpha=1.0, label=f'{final_value:.2f}')
        ax.legend()
        pdf.savefig(fig)
        plt.close(fig)

        # Calculating R2 Score for the test set and doing scatter plot
        r2 = r2_score(list(e['out/true'].values()), list(e['out/pred'].values()))
        fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(8, 8))
        fig.suptitle(f'Scatter Plot - TEST SET\n'
                     f'R2: {r2:.3f}')
        ax.set_xlabel('True Value')
        ax.set_ylabel('Predicted Value')
        value_min = min(*e['out/true'].values(), *e['out/pred'].values())
        value_max = max(*e['out/true'].values(), *e['out/pred'].values())
        for index in e['out/true'].keys():
            ax.scatter(e[f'out/true/{index}'], e[f'out/pred/{index}'],
                       color='green', alpha=0.5)
        ax.plot([value_min, value_max], [value_min, value_max], color='black', alpha=0.5)
        pdf.savefig(fig)
        plt.close(fig)




import os
import pathlib
import typing as t

import tensorflow as tf
import tensorflow.keras as ks
import numpy as np
from pycomex.util import Skippable
from pycomex.experiment import SubExperiment
from kgcnn.data.utils import ragged_tensor_from_nested_numpy

from graph_attention_student.models import Megan
from graph_attention_student.training import NoLoss, mse
from graph_attention_student.training import LogProgressCallback


# == DATASET PARAMETERS ==
# For MEGAN we can use dual channel explanations and thus we can also compare with the dual channel
# ground truth explanations
NODE_IMPORTANCES_KEY: t.Optional[str] = 'node_importances_2'
EDGE_IMPORTANCES_KEY: t.Optional[str] = 'edge_importances_2'


# == MODEL PARAMETERS ==
# These paremeters can be used to configure the model

MODEL_NAME: str = 'MEGAN'

# This list defines how many graph convolutional layers to configure the network with. Every element adds
# one layer. The numbers themselves are the hidden units to be used for those layers.
UNITS: t.List[int] = [32, 32, 32]

# The dropout rate which is applied after EVERY graph convolutional layer of the network. Especially for
# large networks (>20 hidden units and multiple importance channels, this dropout proves very useful)
DROPOUT_RATE: float = 0.0

# This is the weight of the additional explanation-only step which is being applied to the network.
# This explanation only step is important to develop actually interpretable explanations. Refer to the
# paper for more details about this.
IMPORTANCE_FACTOR: float = 1.0

# This is another hyperparameter of the explanation only train step. Usually should be between 1 and 10
IMPORTANCE_MULTIPLIER: float = 5

# This is the number of explanation channels that are generated by the model. This is also the number of
# attention heads used in the graph convolutional layers of the network. So to get a "wider" network, this
# parameter can also be increased. However, note that the value can only be != 2 if importance factor is
# set to exactly 0.0!
IMPORTANCE_CHANNELS: int = 2

# The coefficient value of the explanation sparsity regularization that is applied to the network. Higher
# values should lead to sparser explanations.
SPARSITY_FACTOR: float = 5.0

# If this is true, the results of the individual attention heads will be concatenated, otherwise the
# average will be used to combine the results.
CONCAT_HEADS: bool = False

# We need to supply the range of possible target values and a reference target value a priori to the
# network for the regression case. The regression limits should be as complete as possible. The reference
# does not have to be in the middle, changing it will influence the development of the explanations quite
# a bit.
REGRESSION_REFERENCE = 0
REGRESSION_LIMITS = [[-3, 3]]

CHANNEL_DIRECTIONS = {
    0: -1,
    1: +1,
}

# At the tail end of the network there is a MLP, which does the final prediction. This list defines the
# depth of that MLP. Each element adds a layer and the values themselves dictate the hidden values.
# In the regression case a final entry with "1" hidden layer as the output layer will implicitly be
# added.
FINAL_UNITS = [60, 20, 1]

# This dropout is applied after EVERY layer in the final MLP. Using this should not be necessary
FINAL_DROPOUT = 0.0

# == TRAINING PARAMETERS ==
EPOCHS = 250
OPTIMIZER_CB = lambda: ks.optimizers.Nadam(learning_rate=0.01)


# == EVALUATION PARAMETERS ==
IMPORTANCE_CHANNEL_LABELS = ['negative', 'positive']


# == EXPERIMENT PARAMETERS ==
PATH = pathlib.Path(__file__).parent.absolute()
EXPERIMENT_PATH = os.path.join(PATH, 'vgd_single.py')
BASE_PATH = PATH
NAMESPACE = 'results/vgd_single_megan'
DEBUG = True
with Skippable(), (se := SubExperiment(EXPERIMENT_PATH, BASE_PATH, NAMESPACE, globals())):

    @se.hook('create_model')
    def create_model(e):
        e.info('MEGAN')
        model = Megan(
            units=e.parameters['UNITS'],
            dropout_rate=e.parameters['DROPOUT_RATE'],
            importance_factor=e.parameters['IMPORTANCE_FACTOR'],
            importance_multiplier=e.parameters['IMPORTANCE_MULTIPLIER'],
            importance_channels=e.parameters['IMPORTANCE_CHANNELS'],
            sparsity_factor=e.parameters['SPARSITY_FACTOR'],
            regression_reference=e.parameters['REGRESSION_REFERENCE'],
            regression_limits=e.parameters['REGRESSION_LIMITS'],
            final_units=e.parameters['FINAL_UNITS'],
            final_dropout_rate=e.parameters['FINAL_DROPOUT'],
            use_graph_attributes=False,
            use_edge_features=True,
            concat_heads=e.parameters['CONCAT_HEADS']
        )
        model.compile(
            loss=[
                mse,
                NoLoss(),
                NoLoss(),
            ],
            loss_weights=[
                1, 0, 0
            ],
            metrics=[mse],
            optimizer=e.parameters['OPTIMIZER_CB'](),
            run_eagerly=False
        )
        return model

    @se.hook('fit_model')
    def fit_model(e, model, x_train, y_train, x_test, y_test):
        history = model.fit(
            x_train,
            y_train,
            batch_size=e.parameters['BATCH_SIZE'],
            epochs=e.parameters['EPOCHS'],
            validation_data=(x_test, y_test),
            validation_freq=1,
            callbacks=[
                LogProgressCallback(
                    logger=e.logger,
                    epoch_step=5,
                    identifier='val_output_1_mean_squared_error'
                )
            ],
            verbose=0
        )
        return history.history


    @se.hook('query_model')
    def query_model(e, model, x, y, include_importances: bool = True):
        e.info('querying the model...')
        out_pred, ni_pred, ei_pred = model(x)

        if include_importances:
            return out_pred, ni_pred, ei_pred
        else:
            return out_pred


    @se.hook('calculate_fidelity')
    def calculate_fidelity(e, model, indices_true, x_true, y_true, out_pred, ni_pred, ei_pred):
        rep = e['rep']

        # ~ fidelity
        # For each importance channel we construct a mask which only masks out that very channel
        # and then we query the model using that mask, which effectively means that this channel
        # has absolutely no effect on the prediction. We record the outputs generated by these
        # masked predictions and then afterwards calculate the fidelity from that.
        for k in range(IMPORTANCE_CHANNELS):
            # First of all we need to construct the masks
            masks = []
            for ni in ni_pred:
                mask = np.ones_like(ni)
                mask[:, k] = 0
                masks.append(mask)

            masks_tensor = ragged_tensor_from_nested_numpy(masks)
            out_masked, _, _ = [v.numpy() for v in
                                model(x_true, node_importances_mask=masks_tensor)]

            for c, index in enumerate(indices_true):
                e[f'out/pred/{rep}/{index}'] = float(out_pred[c][0])
                e[f"out/masked/{rep}/{index}/{k}"] = float(out_masked[c][0])

        fidelities = []
        for index in indices_true:

            fidelity = 0
            for k in range(IMPORTANCE_CHANNELS):
                out = e[f"out/pred/{rep}/{index}"]
                out_masked = e[f"out/masked/{rep}/{index}/{k}"]
                fidelity += e.parameters['CHANNEL_DIRECTIONS'][k] * (out - out_masked)

            fidelities.append(fidelity)

        return fidelities




import typing as t

import tensorflow as tf
import tensorflow.keras as ks
from tensorflow.python.keras.engine import compile_utils
from kgcnn.data.utils import ragged_tensor_from_nested_numpy
from kgcnn.layers.modules import GraphBaseLayer
from kgcnn.layers.modules import LazyConcatenate, LazyAverage
from kgcnn.layers.modules import DenseEmbedding, ActivationEmbedding, DropoutEmbedding
from kgcnn.layers.pooling import PoolingLocalEdges
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.layers.pooling import PoolingWeightedNodes
from graph_attention_student.layers import ExplanationSparsityRegularization
from graph_attention_student.layers import MultiHeadGATV2Layer
from graph_attention_student.training import mae
from graph_attention_student.training import bce
from graph_attention_student.training import shifted_sigmoid


class Megan(ks.models.Model):
    """
    MEGAN: Multi Explanation Graph Attention Network
    This model currently supports graph regression and graph classification problems. It was mainly designed
    with a focus on explainable AI (XAI). Along the main prediction, this model is able to output multiple
    attention-based explanations for that prediction. More specifically, the model outputs node and edge
    attributional explanations (assigning [0, 1] values to ever node / edge of the input graph) in K
    separate explanation channels, where K can be chosen as an independent model parameter.
    """

    def __init__(self,
                 # convolutional network related arguments
                 units: t.List[int],
                 activation: str = "kgcnn>leaky_relu",
                 use_bias: bool = True,
                 dropout_rate: float = 0.0,
                 use_edge_features: bool = True,
                 # node/edge importance related arguments
                 importance_units: t.List[int] = [],
                 importance_channels: int = 2,
                 importance_activation: str = "sigmoid",  # do not change
                 importance_dropout_rate: float = 0.0,  # do not change
                 importance_factor: float = 0.0,
                 importance_multiplier: float = 10.0,
                 importance_transformations: t.Optional[t.List[ks.layers.Layer]] = None,
                 sparsity_factor: float = 0.0,
                 concat_heads: bool = True,
                 separate_explanation_step: bool = False,
                 # mlp tail end related arguments
                 final_units: t.List[int] = [1],
                 final_dropout_rate: float = 0.0,
                 final_activation: str = 'linear',
                 final_pooling: str = 'sum',
                 regression_limits: t.Optional[t.Tuple[float, float]] = None,
                 regression_bins: t.Optional[t.List[t.Tuple[float, float]]] = None,
                 regression_reference: t.Optional[float] = None,
                 return_importances: bool = True,
                 use_graph_attributes: bool = False,
                 **kwargs):
        """
        Args:
            units: A list of ints where each element configures an additional attention layer. The numeric
                value determines the number of hidden units to be used in the attention heads of that layer
            activation: The activation function to be used within the attention layers of the network
            use_bias: Whether the layers of the network should use bias weights at all
            dropout_rate: The dropout rate to be applied after *each* of the attention layers of the network.
            use_edge_features: Whether edge features should be used. Generally the network supports the
                usage of edge features, but if the input data does not contain edge features, this should be
                set to False.
            importance_units: A list of ints where each element configures another dense layer in the
                subnetwork that produces the node importance tensor from the main node embeddings. The
                numeric value determines the number of hidden units in that layer.
            importance_channels: The int number of explanation channels to be produced by the network. This
                is the value referred to as "K". Note that this will also determine the number of attention
                heads used within the attention subnetwork.
            importance_factor: The weight of the explanation-only train step. If this is set to exactly
                zero then the explanation train step will not be executed at all (less computationally
                expensive)
            importance_multiplier: An additional hyperparameter of the explanation-only train step. This
                is essentially the scaling factor that is applied to the values of the dataset such that
                the target values can reasonably be approximated by a sum of [0, 1] importance values.
            sparsity_factor: The coefficient for the sparsity regularization of the node importance
                tensor.
            concat_heads: Whether to concat the heads of the attention subnetwork. The default is True. In
                that case the output of each individual attention head is concatenated and the concatenated
                vector is then used as the input of the next attention layer's heads. If this is False, the
                vectors are average pooled instead.
            final_units: A list of ints where each element configures another dense layer in the MLP
                at the tail end of the network. The numeric value determines the number of the hidden units
                in that layer. Note that the final element in this list has to be the same as the dimension
                to be expected for the samples of the training dataset!
            final_dropout_rate: The dropout rate to be applied after *every* layer of the final MLP.
            final_activation: The activation to be applied at the very last layer of the MLP to produce the
                actual output of the network.
            final_pooling: The pooling method to be used during the global pooling phase in the network.
            regression_limits: A tuple where the first value is the lower limit for the expected value range
                of the regression task and teh second value the upper limit.
            regression_reference: A reference value which is inside the range of expected values (best if
                it was in the middle, but does not have to). Choosing different references will result
                in different explanations.
            return_importances: Whether the importance / explanation tensors should be returned as an output
                of the model. If this is True, the output of the model will be a 3-tuple:
                (output, node importances, edge importances), otherwise it is just the output itself
        """
        ks.models.Model.__init__(self, **kwargs)
        self.units = units
        self.activation = activation
        self.use_bias = use_bias
        self.dropout_rate = dropout_rate
        self.use_edge_features = use_edge_features
        self.importance_units = importance_units
        self.importance_channels = importance_channels
        self.importance_activation = importance_activation
        self.importance_dropout_rate = importance_dropout_rate
        self.importance_factor = importance_factor
        self.importance_multiplier = importance_multiplier
        self.importance_transformations = importance_transformations
        self.sparsity_factor = sparsity_factor
        self.concat_heads = concat_heads
        self.final_units = final_units
        self.final_dropout_rate = final_dropout_rate
        self.final_activation = final_activation
        self.final_pooling = final_pooling
        self.regression_limits = regression_limits
        self.regression_reference = regression_reference
        self.regression_bins = regression_bins
        self.return_importances = return_importances
        self.separate_explanation_step = separate_explanation_step
        self.use_graph_attributes = use_graph_attributes

        # ~ MAIN CONVOLUTIONAL / ATTENTION LAYERS
        self.attention_layers: t.List[GraphBaseLayer] = []
        for u in self.units:
            lay = MultiHeadGATV2Layer(
                units=u,
                num_heads=self.importance_channels,
                use_edge_features=self.use_edge_features,
                activation=self.activation,
                use_bias=self.use_bias,
                has_self_loops=True,
                concat_heads=self.concat_heads
            )
            self.attention_layers.append(lay)

        self.lay_dropout = DropoutEmbedding(rate=self.dropout_rate)
        self.lay_sparsity = ExplanationSparsityRegularization(factor=self.sparsity_factor)

        # ~ EDGE IMPORTANCES
        self.lay_act_importance = ActivationEmbedding(activation=self.importance_activation)
        self.lay_concat_alphas = LazyConcatenate(axis=-1)

        self.lay_pool_edges_in = PoolingLocalEdges(pooling_method='mean', pooling_index=0)
        self.lay_pool_edges_out = PoolingLocalEdges(pooling_method='mean', pooling_index=1)
        self.lay_average = LazyAverage()

        # ~ NODE IMPORTANCES
        self.node_importance_units = importance_units + [self.importance_channels]
        self.node_importance_acts = ['kgcnn>leaky_relu' for _ in importance_units] + ['linear']
        self.node_importance_layers = []
        for u, act in zip(self.node_importance_units, self.node_importance_acts):
            lay = DenseEmbedding(
                units=u,
                activation=act,
                use_bias=use_bias
            )
            self.node_importance_layers.append(lay)

        # ~ OUTPUT / MLP TAIL END
        self.lay_pool_out = PoolingNodes(pooling_method=self.final_pooling)
        self.lay_concat_out = LazyConcatenate(axis=-1)
        self.lay_final_dropout = DropoutEmbedding(rate=self.final_dropout_rate)

        self.final_acts = ['kgcnn>leaky_relu' for _ in self.final_units]
        self.final_acts[-1] = self.final_activation
        self.final_biases = [True for _ in self.final_units]
        self.final_biases[-1] = True
        self.final_layers = []
        for u, act, bias in zip(self.final_units, self.final_acts, self.final_biases):
            lay = DenseEmbedding(
                units=u,
                activation=act,
                use_bias=use_bias
            )
            self.final_layers.append(lay)

        # ~ EXPLANATION ONLY TRAIN STEP
        self.bce_loss = ks.losses.BinaryCrossentropy()
        self.compiled_classification_loss = compile_utils.LossesContainer(self.bce_loss)

        self.mse_loss = ks.losses.MeanSquaredError()
        self.mae_loss = ks.losses.MeanAbsoluteError()
        self.compiled_regression_loss = compile_utils.LossesContainer(mae)

        # If regression_limits have been supplied, we interprete this as the intent to perform explanation
        # co-training for a regression dataset.
        # So the content of this if condition makes sure to perform the necessary pre-processing steps
        # for this case.
        if self.regression_limits is not None:
            # first of all we do some simple assertions to handle some common error cases
            assert regression_reference is not None, ('You have to supply a "regression_reference" value for '
                                                      'explanation co-training!')

            # This is the first and simpler case for regression explanation co-training: In this case the
            # regression reference value is only a single value. In that case, there is only one target
            # value that is supposed to be regressed. The alternative would be that it is a list in which
            # case it would have to have as many elements as target values to be predicted.
            # However in this case we convert it into a list as well to be able to treat everything from
            # this point on as the multi-value case guaranteed.
            if isinstance(regression_reference, (int, float)):
                self.regression_reference = [regression_reference]

            num_references = len(self.regression_reference)
            num_limits = len(self.regression_limits)
            assert num_references * 2 == importance_channels, (
                f'for explanation co-training, the number of regression_references (currently {num_references}) has '
                f'to be exactly half the number of importance channels (currently {importance_channels})!'
            )
            assert num_references == final_units[-1], (
                f'For explanation co-training, the number of regression_references (currently {num_references}) has '
                f'to be exactly the same as the final unit count in the MLP tail end (currently {final_units[-1]})'
            )
            assert num_references == num_limits, (
                f'For explanation co-training, the number of regression_references (currently {num_references}) has '
                f'to be exactly the same as the number of regression_limits intervals (currently {num_limits})'
            )

    def get_config(self):
        config = super(Megan, self).get_config()
        config.update({
            "units": self.units,
            "activation": self.activation,
            "use_bias": self.use_bias,
            "dropout_rate": self.dropout_rate,
            "use_edge_features": self.use_edge_features,
            "importance_units": self.importance_units,
            "importance_channels": self.importance_channels,
            "importance_activation": self.importance_activation,
            "importance_dropout_rate": self.importance_dropout_rate,
            "importance_factor": self.importance_factor,
            "importance_multiplier": self.importance_multiplier,
            "sparsity_factor": self.sparsity_factor,
            "concat_heads": self.concat_heads,
            "final_units": self.final_units,
            "final_dropout_rate": self.final_dropout_rate,
            "final_activation": self.final_activation,
            "final_pooling": self.final_pooling,
            "regression_limits": self.regression_limits,
            "regression_reference": self.regression_reference,
            "return_importances": self.return_importances
        })

        return config
    @property
    def doing_regression(self) -> bool:
        return self.regression_limits is not None

    def call(self,
             inputs,
             training: bool = False,
             return_importances: bool = False,
             node_importances_mask: t.Optional[tf.RaggedTensor] = None,
             **kwargs):
        """
        Forward pass of the model.

        **Shape Explanations:** All shapes in brackets [] are ragged dimensions!

        - V: Num nodes in the graph
        - E: Num edges in the graph
        - N: Num feature values per node
        - M: NUm feature values per edge
        - H: Num feature values per graph
        - B: Num graphs in a batch
        - K: Num importance (explanation) channels configured in the constructor
        """

        # 17.11.2022
        # Added support for global graph attributes. If the corresponding flag is set in the constructor of
        # the model then it is expected that the input tuple consists of 4 elements instead of the usual
        # 3 elements, where the fourth element is the vector of the graph attributes.
        # We can't use these graph attributes right away, but later on we will simply append them to the
        # vector which enters the MLP tail end.

        # node_input: ([B], [V], N)
        # edge_input: ([B], [E], M)
        # edge_index_input: ([B], [E], 2)
        # graph_input: ([B], H)
        if self.use_graph_attributes:
            node_input, edge_input, edge_index_input, graph_input = inputs
        else:
            node_input, edge_input, edge_index_input = inputs
            graph_input = None

        # First of all we apply all the graph convolutional / attention layers. Each of those layers outputs
        # the attention logits alpha additional to the node embeddings. We collect all the attention logits
        # in a list so that we can later sum them all up.
        alphas = []
        x = node_input
        for lay in self.attention_layers:
            # x: ([batch], [N], F)
            # alpha: ([batch], [M], K, 1)
            x, alpha = lay([x, edge_input, edge_index_input])
            if training:
                x = self.lay_dropout(x, training=training)

            alphas.append(alpha)

        # We sum up all the individual layers attention logit tensors and the edge importances are directly
        # calculated by applying a sigmoid on that sum.
        alphas = self.lay_concat_alphas(alphas)
        edge_importances = tf.reduce_sum(alphas, axis=-1, keepdims=False)
        edge_importances = self.lay_act_importance(edge_importances)

        # Part of the final node importance tensor is actually the pooled edge importances, so that is what
        # we are doing here. The caveat here is that we assume undirected edges as two directed edges in
        # opposing direction. To now achieve a symmetric pooling of these edges we have to pool in both
        # directions and then use the average of both.
        pooled_edges_in = self.lay_pool_edges_in([node_input, edge_importances, edge_index_input])
        pooled_edges_out = self.lay_pool_edges_out([node_input, edge_importances, edge_index_input])
        pooled_edges = self.lay_average([pooled_edges_out, pooled_edges_in])

        node_importances_tilde = x
        for lay in self.node_importance_layers:
            node_importances_tilde = lay(node_importances_tilde)

        node_importances_tilde = self.lay_act_importance(node_importances_tilde)

        node_importances = node_importances_tilde * pooled_edges
        self.lay_sparsity(node_importances)

        # ~ Applying the node importance mask
        # "node_importances_mask" is supposed to be a ragged tensor of the exact same dimensions as the
        # node importances, containing binary values 0 or 1, which are then used as a multiplicative mask
        # to modify the actual node importances before the global pooling step.
        # The main use case of this feature is to completely mask out certain channels to see how that
        # the missing channels (linked to a certain explanation / interpretation) affect the outcome of
        # the MLP tail end.
        if node_importances_mask is not None:
            node_importances_mask = tf.cast(node_importances_mask, tf.float32)
            node_importances = node_importances * node_importances_mask

        # Here we apply the global pooling. It is important to note that we do K separate pooling operations
        # were each time we use the same node embeddings x but a different slice of the node importances as
        # the weights! We concatenate all the individual results in the end.
        outs = []
        n = self.final_units[-1]
        for k in range(self.importance_channels):
            node_importance_slice = tf.expand_dims(node_importances[:, :, k], axis=-1)
            masked_embeddings = x * node_importance_slice

            # 26.03.2023
            # Optionally, if given we apply an additional non-linear transformation in the form of an
            # additional layer on each of the masked node embeddings separately.
            if self.importance_transformations is not None:
                lay_transform = self.importance_transformations[k]
                masked_embeddings = lay_transform(masked_embeddings)

            out = self.lay_pool_out(masked_embeddings)
            # out = self.lay_pool_out(x[:, :, k*n:(k+1)*n] * node_importance_slice)

            outs.append(out)

        # out: ([B], N*K²)
        out = self.lay_concat_out(outs)

        # At this point, after the global pooling of the node embeddings, we can append the global graph
        # attributes, should those exist
        if self.use_graph_attributes:
            out = self.lay_concat_out([out, graph_input])

        # Now "out" is a graph embedding vector of known dimension so we can simply apply the normal dense
        # mlp to get the final output value.
        num_final_layers = len(self.final_layers)
        for c, lay in enumerate(self.final_layers):
            out = lay(out)
            if training and c < num_final_layers - 2:
                out = self.lay_final_dropout(out, training=training)

        if self.doing_regression:
            reference = tf.ones_like(out) * tf.constant(self.regression_reference, dtype=tf.float32)
            out = out + reference

        # Usually, the node and edge importance tensors would be direct outputs of the model as well, but
        # we need the option to just return the output alone to be compatible with the standard model
        # evaluation pipeline already implemented in the library.
        if self.return_importances or return_importances:
            return out, node_importances, edge_importances
        else:
            return out

    def regression_augmentation(self,
                                out_true):
        samples = []
        masks = []
        for i, (regression_reference, regression_limits) in enumerate(zip(self.regression_reference,
                                                                          self.regression_limits)):

            regression_width = abs(regression_limits[1] - regression_limits[0])
            values = tf.expand_dims(out_true[:, i], axis=-1)
            center_distances = tf.abs(values - regression_reference)
            center_distances = (center_distances * self.importance_multiplier) / (0.5 * regression_width)

            # So we need two things: a "samples" tensor and a "mask" tensor. We are going to use the samples
            # tensor as the actual ground truth which acts as the regression target during the explanation
            # train step. The binary values of the mask will determine at which positions a loss should
            # actually be calculated for both of the channels

            # The "lower" part is all the samples which have a target value below the reference value.
            lo_mask = tf.where(values < regression_reference, 1.0, 0.0)
            # The "higher" part is all the samples above reference
            hi_mask = tf.where(values > regression_reference, 1.0, 0.0)

            samples += [center_distances, center_distances]
            masks += [lo_mask, hi_mask]

        return (
            tf.concat(samples, axis=-1),
            tf.concat(masks, axis=-1)
        )

    def train_step_explanation(self, x, y,
                               update_weights: bool = True):

        if self.return_importances:
            out_true, _, _ = y
        else:
            out_true = y

        exp_loss = 0
        with tf.GradientTape() as tape:

            y_pred = self(x, training=True, return_importances=True)
            out_pred, ni_pred, ei_pred = y_pred

            # ~ explanation loss
            # First of all we need to assemble the approximated model output, which is simply calculated
            # by applying a global pooling operation on the corresponding slice of the node importances.
            # So for each slice (each importance channel) we get a single value, which we then
            # concatenate into an output vector with K dimensions.
            outs = []
            for k in range(self.importance_channels):
                node_importances_slice = tf.expand_dims(ni_pred[:, :, k], axis=-1)
                out = self.lay_pool_out(node_importances_slice)

                outs.append(out)

            # outs: ([batch], K)
            outs = self.lay_concat_out(outs)

            if self.doing_regression:
                out_true, mask = self.regression_augmentation(out_true)
                out_pred = outs
                exp_loss = self.importance_channels * self.compiled_regression_loss(out_true * mask,
                                                                                    out_pred * mask)

            else:
                # out_pred = ks.backend.sigmoid(outs)
                out_pred = shifted_sigmoid(outs, shift=self.importance_multiplier, multiplier=2)
                exp_loss = self.compiled_classification_loss(out_true, out_pred * out_true)

            exp_loss *= self.importance_factor

        # Compute gradients
        trainable_vars = self.trainable_variables
        exp_gradients = tape.gradient(exp_loss, trainable_vars)

        # Update weights
        if update_weights:
            self.optimizer.apply_gradients(zip(exp_gradients, trainable_vars))

        return {'exp_loss': exp_loss}

    def train_step(self, data):
        if len(data) == 3:
            x, y, sample_weight = data
        else:
            sample_weight = None
            x, y = data

        # The "train_step_explanation" method will execute the entire explanation train step once. This
        # mainly involves creating an approximate solution of the original regression / classification
        # problem using ONLY the node importances of the different channels and then performing one
        # complete weight update based on the corresponding loss.
        if self.importance_factor != 0 and self.separate_explanation_step:
            exp_metrics = self.train_step_explanation(x, y)
        else:
            exp_metrics = {}

        exp_loss = 0
        with tf.GradientTape() as tape:

            out_true, ni_true, ei_true = y
            out_pred, ni_pred, ei_pred = self(x, training=True, return_importances=True)
            loss = self.compiled_loss(
                [out_true, ni_true, ei_true],
                [out_pred, ni_pred, ei_pred],
                sample_weight=sample_weight,
                regularization_losses=self.losses,
            )

            if self.importance_factor != 0 and not self.separate_explanation_step:
                # ~ explanation loss
                # First of all we need to assemble the approximated model output, which is simply calculated
                # by applying a global pooling operation on the corresponding slice of the node importances.
                # So for each slice (each importance channel) we get a single value, which we then
                # concatenate into an output vector with K dimensions.
                outs = []
                for k in range(self.importance_channels):
                    node_importances_slice = tf.expand_dims(ni_pred[:, :, k], axis=-1)
                    out = self.lay_pool_out(node_importances_slice)

                    outs.append(out)

                # outs: ([batch], K)
                outs = self.lay_concat_out(outs)

                if self.doing_regression:
                    _out_true, mask = self.regression_augmentation(out_true)
                    _out_pred = outs
                    exp_loss = self.compiled_regression_loss(_out_true * mask,
                                                             _out_pred * mask)

                else:
                    # out_pred = ks.backend.sigmoid(outs)
                    #_out_pred = shifted_sigmoid(outs, multiplier=self.importance_multiplier)
                    _out_pred = shifted_sigmoid(
                        outs,
                        shift=self.importance_multiplier,
                        multiplier=1
                    ) * out_true
                    exp_loss = self.compiled_classification_loss(out_true, _out_pred)

                exp_loss *= self.importance_factor
                exp_metrics['exp_loss'] = exp_loss
                loss += exp_loss

        trainable_vars = self.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)

        self.optimizer.apply_gradients(zip(gradients, trainable_vars))

        self.compiled_metrics.update_state(
            y,
            out_pred if not self.return_importances else [out_pred, ni_pred, ei_pred],
            sample_weight=sample_weight
        )

        # Return a dict mapping metric names to current value.
        # Note that it will include the loss (tracked in self.metrics).
        return {
            **{m.name: m.result() for m in self.metrics},
            **exp_metrics
        }


import time
import types
import logging
from typing import List, Dict, Optional, Callable
from collections import defaultdict

import tensorflow as tf
import tensorflow.keras as ks


# == CUSTOM ACTIVATION FUNCTIONS ============================================================================
# ===========================================================================================================

def shifted_sigmoid(x, multiplier: float = 1, shift: float = 10) -> float:
    return ks.backend.sigmoid(multiplier * (x - shift))


# == TRAINING SEGMENTATION ==================================================================================
# ===========================================================================================================


class RecompilableMixin:

    def __init__(self):
        self.compile_args: Optional[list] = None
        self.compile_kwargs: Optional[dict] = None

    def compile(self, *args, **kwargs):
        ks.models.Model.compile(self, *args, **kwargs)
        self.compile_args = args
        self.compile_kwargs = kwargs

    def recompile(self):
        self.compile(*self.compile_args, **self.compile_kwargs)


class SegmentedFitProcess:

    def __init__(self,
                 model: ks.models.Model,
                 fit_kwargs: dict):
        self.model = model
        self.fit_kwargs = fit_kwargs

        self.total_epochs = fit_kwargs['epochs']
        self.callbacks = dict()
        self.callbacks[self.total_epochs] = lambda model, history: None
        self.current_epoch = 0

    def __call__(self):
        """
        This method actually executes the various model.fit() operations which are part of the overall
        segmented process. It then returns a ks History object which contains the *overall* history which
        consists of the merged histories of all the individual processes.

        :return:
        """
        hists = []

        # self.callbacks: Dict[int, Callable] is a dictionary whose keys are integer epoch numbers which
        # are generally between 0 and the total number of epochs defined for the entire fit process.
        # The corresponding values are callbacks which are applied *on the model* at the point at which the
        # overall process hits that epoch.
        for epoch, cb in sorted(self.callbacks.items(), key=lambda i: i[0]):
            # The next fit process has to have as many epochs as starting from the current one and until the
            # next callback has to be applied.
            epoch_diff = epoch - self.current_epoch

            # For each individual fit process we can just reuse the majority of the fit kwargs, but we need
            # to replace the epoch count with the local count only until the next callback is due.
            fit_kwargs = self.fit_kwargs.copy()
            fit_kwargs['epochs'] = epoch_diff

            # We train the model and then after it is done we apply the callback.
            hist = self.model.fit(**fit_kwargs)
            hists.append(hist)
            cb(
                model=self.model,
                history=hist
            )

            self.current_epoch += epoch_diff

        merged_hist = self.merge_histories(hists)
        return merged_hist

    def merge_histories(self, hists: List[ks.callbacks.History]):
        merged_hist = ks.callbacks.History()

        merged_history = defaultdict(list)
        for hist in hists:
            for key, value in hist.history.items():
                if isinstance(value, list):
                    merged_history[key] += value

        merged_hist.history = merged_history
        return merged_hist

    def add_callback(self,
                     epochs: int,
                     callback: Callable[[ks.models.Model, ks.callbacks.History], None]
                     ) -> None:
        self.callbacks[epochs] = callback

# This section is a lot of black magic
# ------------------------------------

def epochs_once(epoch: int):

    def register(self, instance, fit_process, fit_kwargs):
        # https://stackoverflow.com/questions/1015307/python-bind-an-unbound-method
        bound_method = self.__get__(instance, instance.__class__)
        fit_process.add_callback(epoch, bound_method)

    def local(func: Callable):
        method = types.MethodType(register, func)
        setattr(func, 'register', method)
        setattr(func, 'is_epoch_callback', True)

        return func

    return local


class FitManager:

    def __init__(self):
        pass

    def __call__(self,
                 model: ks.models.Model,
                 fit_kwargs: dict) -> SegmentedFitProcess:
        fit_process = SegmentedFitProcess(model, fit_kwargs)
        # self.register_callbacks(fit_process, fit_kwargs, model)
        callbacks = self.gather_callbacks()
        for callback in callbacks:
            callback.register(self, fit_process, fit_kwargs)

        return fit_process

    def gather_callbacks(self) -> List[Callable]:
        callbacks = []
        for name in dir(self):
            attribute = getattr(self, name)
            if callable(attribute) and hasattr(attribute, 'is_epoch_callback'):
                callbacks.append(attribute)

        return callbacks

    def register_callbacks(self,
                           fit_process: SegmentedFitProcess,
                           fit_kwargs: dict,
                           model: ks.models.Model) -> None:
        return


class LockExplanationManager(FitManager):

    def __init__(self,
                 epoch_threshold: int):
        FitManager.__init__(self)
        self.epoch_threshold = epoch_threshold

    def register_callbacks(self, fit_process, fit_kwargs, model):
        fit_process.add_callback(self.epoch_threshold, self.lock_model_explanation_parameters)

    def lock_model_explanation_parameters(self, model, history):
        model.lock_explanation_parameters()
        model.recompile()


class NoLoss(ks.losses.Loss):

    def __init__(self,
                 *args,
                 name='no_loss',
                 **kwargs):
        super(NoLoss, self).__init__(*args, name=name, **kwargs)

    def call(self, y_true, y_pred):
        return 0


def bce(y_true, y_pred):
    y_true = tf.cast(y_true, tf.float32)
    loss = y_true * tf.math.log(y_pred + 1e-7)
    loss += (1 - y_true) * tf.math.log(1 - y_pred + 1e-7)
    loss = -loss
    return tf.reduce_mean(loss)


def mae(y_true, y_pred):
    y_true = tf.cast(y_true, tf.float32)
    nan_multiplier = tf.where(tf.math.is_nan(y_true), 0., 1.)

    y_true = tf.where(tf.math.is_nan(y_true), 0., y_true)
    loss = tf.abs(y_true - y_pred)

    loss *= nan_multiplier
    loss = tf.reduce_sum(loss, axis=-1)

    return tf.reduce_mean(loss)


mae.name = 'mean_absolute_error'


def mse(y_true, y_pred):
    nan_multiplier = tf.where(tf.math.is_nan(y_true), 0., 1.)

    y_true = tf.cast(y_true, tf.float32)
    y_true = tf.where(tf.math.is_nan(y_true), 0., y_true)
    loss = tf.square(y_true - y_pred)

    loss *= nan_multiplier
    loss = tf.reduce_sum(loss, axis=-1)
    # print(loss)

    return tf.reduce_mean(loss)


mse.name = 'mean_squared_error'


class ExplanationLoss(ks.losses.Loss):

    def __init__(self,
                 loss_function: ks.losses.Loss = bce,
                 mask_empty_explanations: bool = False,
                 reduce: bool = False,
                 factor: float = 1):
        ks.losses.Loss.__init__(self)

        self.loss_function = loss_function
        self.mask_empty_explanations = mask_empty_explanations
        self.factor = factor
        self.reduce = reduce
        self.name = 'explanation_loss'

    def call(self, y_true, y_pred):
        loss = self.loss_function(y_true, y_pred)

        if self.mask_empty_explanations:
            mask = tf.cast(tf.reduce_max(y_true, axis=-1) > 0, dtype=tf.float32)
            loss *= mask

        if self.reduce:
            loss = tf.reduce_mean(loss, axis=-1)

        return self.factor * loss


class EpochCounterCallback(ks.callbacks.Callback):

    def __init__(self):
        super(EpochCounterCallback, self).__init__()
        self.epoch = 0

    def on_epoch_begin(self, epoch, logs=None):
        self.epoch += 1


class LogProgressCallback(EpochCounterCallback):

    def __init__(self,
                 logger: logging.Logger,
                 identifier: str,
                 epoch_step: int):
        super(LogProgressCallback, self).__init__()

        self.logger = logger
        self.identifier = identifier
        self.epoch_step = epoch_step

        self.start_time = time.time()
        self.elapsed_time = 0

    def on_epoch_end(self, epoch, logs=None):
        if (self.epoch % self.epoch_step == 0 or self.epoch == 1) and logs is not None:
            self.elapsed_time = time.time() - self.start_time
            value = logs[self.identifier]

            message_parts = [
                f'   epoch {str(self.epoch):<5}: ',
                f'{self.identifier}={value:.3f} ',
                f'elapsed_time={self.elapsed_time:.1f}s'
            ]
            if 'output_1_loss' in logs:
                message_parts.append(f'output_1_loss={logs["output_1_loss"]:.3f} ')
            if 'output_2_loss' in logs:
                message_parts.append(f'output_2_loss={logs["output_2_loss"]:.3f}')
            if 'output_3_loss' in logs:
                message_parts.append(f'output_3_loss={logs["output_3_loss"]:.3f}')
            if 'loss' in logs:
                message_parts.append(f'total_training_loss={logs["loss"]:.3f} ')
            if 'exp_loss' in logs:
                message_parts.append(f'exp_loss={logs["exp_loss"]:.3f}')

            self.logger.info(' '.join(message_parts))

    def reset(self):
        self.epoch = 0
        self.start_time = time.time()
        self.elapsed_time = 0


import os
import pathlib
import typing as t

import numpy as np
import tensorflow as tf
import tensorflow.keras as ks
from pycomex.experiment import SubExperiment
from pycomex.util import Skippable
from kgcnn.data.utils import ragged_tensor_from_nested_numpy
from graph_attention_student.models import Megan
from graph_attention_student.training import NoLoss
from graph_attention_student.training import ExplanationLoss

PATH = pathlib.Path(__file__).parent.absolute()

# == DATASET PROPERTIES ==
VISUAL_GRAPH_DATASET_PATH: str = os.path.expanduser('~/.visual_graph_datasets/datasets/rb_dual_motifs')
NODE_IMPORTANCES_KEY = 'node_importances_2'
EDGE_IMPORTANCES_KEY = 'edge_importances_2'

# == MODEL PARAMETERS ==
UNITS: t.List[int] = [6, 6, 6]
FINAL_UNITS: t.List[int] = [3, 1]
FINAL_ACTIVATION: str = 'linear'
IMPORTANCE_CHANNELS: int = 2
IMPORTANCE_FACTOR: float = 1.0
IMPORTANCE_MULTIPLIER: float = 2.0
REGRESSION_LIMITS = [[-3, +3]]
REGRESSION_REFERENCE = [0]
CHANNEL_DIRECTIONS = {0: -1, 1: 1}


# == EXPERIMENT PARAMETERS ==
EXPERIMENT_PATH = os.path.join(PATH, 'vgd_regression_supervised.py')
BASE_PATH = PATH
NAMESPACE = 'results/' + os.path.basename(__file__).strip('.py')
DEBUG = True
with Skippable(), (se := SubExperiment(EXPERIMENT_PATH, BASE_PATH, NAMESPACE, globals())):

    @se.hook('model_generator')
    def model_generator(e):
        model = Megan(
            units=UNITS,
            importance_channels=IMPORTANCE_CHANNELS,
            importance_factor=IMPORTANCE_FACTOR,
            importance_multiplier=IMPORTANCE_MULTIPLIER,
            regression_limits=REGRESSION_LIMITS,
            regression_reference=REGRESSION_REFERENCE,
            final_units=FINAL_UNITS,
            final_activation=FINAL_ACTIVATION,
        )
        model.compile(
            loss=[
                ks.losses.MeanSquaredError(),
                NoLoss(),
                NoLoss(),
            ],
            loss_weights=[1, 0, 0],
            metrics=[ks.metrics.MeanSquaredError()],
            optimizer=e.parameters['OPTIMIZER_CB']()
        )

        yield 'megan_2_normal', model

        model = Megan(
            units=UNITS,
            importance_channels=IMPORTANCE_CHANNELS,
            importance_factor=0.0,
            final_units=FINAL_UNITS,
            final_activation=FINAL_ACTIVATION,
        )
        model.compile(
            loss=[
                ks.losses.MeanSquaredError(),
                ExplanationLoss(),
                ExplanationLoss(),
            ],
            loss_weights=[1, 1, 1],
            metrics=[ks.metrics.MeanSquaredError()],
            optimizer=e.parameters['OPTIMIZER_CB']()
        )

        yield 'megan_2_supervised', model

    @se.hook('additional_metrics')
    def calculate_fidelity_ast(e, model, test_indices, x_test, y_test):
        # This metric can only be computed for MEGAN models, so we make sure that this is the case here
        if isinstance(model, Megan):
            e.info('calculating fidelity*, which is specific to multi-channel MEGAN models ...')
            importance_channels = e.parameters['IMPORTANCE_CHANNELS']
            rep = e['rep']
            key = e['key']

            out_pred, ni_pred, ei_pred = [v.numpy() for v in model(x_test)]

            # First of all we need to calculate the masked output deviations for each of the channels
            # and each of the elements in the dataset.
            for i in range(importance_channels):
                channel_direction = e.parameters['CHANNEL_DIRECTIONS'][i]

                # To do that we construct the corresponding masks here. For each channel we construct the
                # mask such that only that channel will be removed from the final pooling operation.
                node_masks = []
                for c, index in enumerate(test_indices):
                    ni = ni_pred[c]
                    node_mask = np.ones_like(ni)
                    node_mask[:, i] = 0
                    node_masks.append(node_mask)

                node_masks_tensor = ragged_tensor_from_nested_numpy(node_masks)
                # Using these tensors we can query the model again to get the modified predictions outputs
                out_mod, _, _ = [v.numpy() for v in model(x_test, node_importances_mask=node_masks_tensor)]
                for c, index in enumerate(test_indices):
                    value = channel_direction * (out_pred[c] - out_mod[c])
                    e[f'fidelity_ast_contribution/{rep}/{key}/{index}/{i}'] = value

            # Now we merge all the individual channel contributions into the final fidelity value
            for c, index in enumerate(test_indices):
                e[f'fidelity_ast/{rep}/{key}/{index}'] = np.sum([
                    e[f'fidelity_ast_contribution/{rep}/{key}/{index}/{i}']
                    for i in range(importance_channels)
                ])

            fidelities = [e[f'fidelity_ast/{rep}/{key}/{index}'] for index in test_indices]
            e.info(f'fidelity_ast results:\n'
                   f' * median: {np.median(fidelities):.3f}\n'
                   f' * mean: {np.mean(fidelities):.3f}\n'
                   f' * std: {np.std(fidelities):.3f}\n')

    # This hook adds the "fidelity_ast" metric results as an additional column to the final latex
    # evaluation table.
    @se.hook('additional_columns')
    def additional_columns(e):

        def fidelity_ast_callback(key, rep):
            return list(e[f'fidelity_ast/{rep}/{key}'].values())

        return [r'$\text{fidelity}^{*}$'], [fidelity_ast_callback]



import os
import pathlib
import typing as t

from pycomex.util import Skippable
from pycomex.experiment import SubExperiment

# == DATASET PARAMETERS ==

VISUAL_GRAPH_DATASET_PATH: str = os.path.expanduser('~/.visual_graph_datasets/datasets/aqsoldb')
USE_DATASET_SPLIT: t.Optional[int] = 0
TRAIN_RATIO: float = 0.8
NUM_EXAMPLES: int = 100
EXAMPLE_INDICES = [7691, 7535, 7175, 4672, 8025, 4735, 7794, 7750]
IMPORTANCE_CHANNELS = 1
NODE_IMPORTANCES_KEY: t.Optional[str] = None
EDGE_IMPORTANCES_KEY: t.Optional[str] = None

USE_NODE_COORDINATES: bool = False
USE_EDGE_LENGTHS: bool = False

# == MODEL PARAMETERS ==
UNITS = [32, 32, 32]
FINAL_UNITS = [32, 16, 1]

# == TRAINING PARAMETERS ==
REPETITIONS = 5
EPOCHS = 250

# == EXPERIMENT PARAMETERS ==
PATH = pathlib.Path(__file__).parent.absolute()
EXPERIMENT_PATH = os.path.join(PATH, 'vgd_single_gnnx.py')
BASE_PATH = os.getcwd()
NAMESPACE = 'results/vgd_single_gnnx_aqsoldb'
DEBUG = True
with Skippable(), (se := SubExperiment(EXPERIMENT_PATH, BASE_PATH, NAMESPACE, globals())):
    pass


"""
This example shows how to train a KGCNN GATv2 for multitask graph regression application based on an
existing "visual graph dataset".

Inherits "vgd_multitask_gnn.py"

**CHANGELOG**

0.1.0 - 18.01.2022 - Initial version

0.2.0 - 20.01.2022 - Extended the model class to implement multi-head attention instead of just using single
attention layers.

"""
import os
import warnings
import pathlib
warnings.filterwarnings("ignore")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import random
import typing as t

import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
import tensorflow.keras as ks
from pycomex.experiment import SubExperiment
from pycomex.util import Skippable
from kgcnn.layers.conv.gat_conv import AttentionHeadGATV2
from kgcnn.layers.modules import LazyConcatenate, DenseEmbedding, DropoutEmbedding, LazyAverage
from kgcnn.layers.pooling import PoolingGlobalEdges

# local
from graph_attention_student.training import mse


# == DATASET PARAMETERS ==

USE_GRAPH_ATTRIBUTES: bool = False
# The GATv2 layers are capable to use edge features. This flag controls if that should be the case or not
# If this flag is False then there won't be any edge information at all but just constant edge weights of
# 1 for every edge.
USE_EDGE_FEATURES: bool = True


# == MODEL PARAMETERS ==

class Gatv2Model(ks.models.Model):

    def __init__(self,
                 units: t.List[int],
                 final_units: t.List[int],
                 pooling_method: str = 'sum',
                 num_heads: int = 2,
                 activation: str = 'kgcnn>leaky_relu',
                 final_pooling: str = 'sum',
                 final_activation: str = 'linear',
                 dropout_rate: float = 0.0,
                 use_graph_attributes: bool = False,
                 use_edge_features: bool = True):
        super(Gatv2Model, self).__init__()
        self.use_graph_attributes = use_graph_attributes

        self.lay_dropout = DropoutEmbedding(rate=dropout_rate)
        self.conv_layers = []
        for k in units:
            heads = []
            for _ in range(num_heads):
                lay = AttentionHeadGATV2(
                    units=k,
                    activation=activation,
                    use_edge_features=use_edge_features,
                )
                heads.append(lay)

            self.conv_layers.append(heads)

        self.lay_combine = LazyAverage()
        self.lay_combine = LazyConcatenate()
        self.lay_pooling = PoolingGlobalEdges(pooling_method=final_pooling)
        self.lay_concat = LazyConcatenate(axis=-1)

        self.final_layers = []
        self.final_activations = ['relu' for _ in final_units]
        self.final_activations[-1] = final_activation
        for k, act in zip(final_units, self.final_activations):
            lay = DenseEmbedding(units=k, activation=act)
            self.final_layers.append(lay)

    def call(self, inputs, training=False):
        if self.use_graph_attributes:
            node_input, edge_input, edge_index_input, graph_input = inputs
        else:
            node_input, edge_input, edge_index_input = inputs

        x = node_input
        for heads in self.conv_layers:

            xs = []
            for lay in heads:
                _x = lay([x, edge_input, edge_index_input])
                xs.append(_x)

            x = self.lay_combine(xs)
            if training:
                x = self.lay_dropout(x)

        final = self.lay_pooling(x)
        if self.use_graph_attributes:
            final = self.lay_concat([final, graph_input])

        for lay in self.final_layers:
            final = lay(final)

        return final


MODEL_NAME = 'GATv2'
UNITS: t.List[int] = [32, 32, 32]
DROPOUT_RATE: float = 0.25
# The number of attention heads to be used in the network.
NUM_HEADS: int = 3

# == TRAINING PARAMETERS ==
EPOCHS: int = 250
LOG_STEP: int = 5

# == EXPERIMENT PARAMETERS ==
PATH = pathlib.Path(__file__).parent.absolute()
EXPERIMENT_PATH = os.path.join(PATH, 'vgd_multitask_gnn.py')
BASE_PATH = os.getcwd()
NAMESPACE = 'results/vgd_multitask_gnn_gatv2'
DEBUG = True
with Skippable(), (se := SubExperiment(EXPERIMENT_PATH, BASE_PATH, NAMESPACE, globals())):

    # Here we need to overwrite this hook in case we want ot use the edge features. In the original
    # experiment, the implementation of this hook removes the edge features from the dataset, which
    # means by overwriting it we would prevent that and allow the edge features to be part of the dataset
    @se.hook('modify_g', replace=USE_EDGE_FEATURES)
    def allow_edge_features(e, g):
        return g

    @se.hook('create_model')
    def create_model(e):
        e.info('creating the GATv2 model...')
        e.info(f' * UNITS: {e.parameters["UNITS"]}')
        e.info(f' * FINAL_UNITS: {e.parameters["FINAL_UNITS"]}')
        e.info(f' * DROPOUT_RATE: {e.parameters["DROPOUT_RATE"]}')

        model = Gatv2Model(
            units=e.parameters['UNITS'],
            final_units=e.parameters['FINAL_UNITS'],
            num_heads=e.parameters['NUM_HEADS'],
            use_edge_features=e.parameters['USE_EDGE_FEATURES'],
            use_graph_attributes=e.parameters['USE_GRAPH_ATTRIBUTES'],
            dropout_rate=e.parameters['DROPOUT_RATE']
        )

        model.compile(
            loss=mse,
            metrics=mse,
            optimizer=ks.optimizers.Adam(learning_rate=e.parameters['LEARNING_RATE']),
            run_eagerly=False
        )

        return model



from graph_attention_student.models.megan import *
from graph_attention_student.models.gradient import *
from graph_attention_student.models.gnes import *
from graph_attention_student.models.gnnx import *



"""
Utility methods
"""
import os
import pathlib
import tempfile
import logging
import subprocess
import shutil
from typing import List, Dict, Any, Tuple, Optional, Union, Callable
from collections import defaultdict

import numpy as np
import jinja2 as j2
import matplotlib.colors as mcolors
from nltk.tokenize import word_tokenize
from scipy.spatial import distance

PATH = pathlib.Path(__file__).parent.absolute()
VERSION_PATH = os.path.join(PATH, 'VERSION')

DATASETS_FOLDER = os.path.join(PATH, 'datasets')
EXAMPLES_FOLDER = os.path.join(PATH, 'examples')
TEMPLATES_FOLDER = os.path.join(PATH, 'templates')
EXPERIMENTS_PATH = os.path.join(PATH, 'experiments')

TEMPLATE_ENV = j2.Environment(
    loader=j2.FileSystemLoader(TEMPLATES_FOLDER),
    autoescape=j2.select_autoescape(),
)
TEMPLATE_ENV.globals.update(**{
    'zip': zip,
    'int': int,
    'enumerate': enumerate
})

NULL_LOGGER = logging.Logger('NULL')
NULL_LOGGER.addHandler(logging.NullHandler())

# These are the default tokens which will be removed from a string during the conversion of text to
# graph
SANITIZE_TOKENS = ['', ',', '.', '\'s', '"', '-', '?', '!', '/', '(', ')', '_', '--', ';', '``']
#SANITIZE_TOKENS += ['the', 'be', 'to', 'of', 'and', 'a', 'an', 'in', 'that', 'it', 'you', 'me', 'i', 'at']


def get_version():
    with open(VERSION_PATH) as file:
        return file.read().replace(' ', '').replace('\n', '')


def update_nested_dict(original: dict, extension: dict):
    result = original.copy()
    for key, value in extension.items():
        if isinstance(value, dict) and key in result:
            result[key] = update_nested_dict(result[key], value)
        else:
            result[key] = value

    return result

# == GENERAL GRAPH OPERATIONS ===============================================================================


def node_adjacency_sliding_window(node_indices: List[int],
                                  window_size: int,
                                  do_self_loops: bool,
                                  ) -> List[List[int]]:
    # First we initialize the adjacency matrix
    node_adjacency = [[0 for _ in node_indices] for _ in node_indices]
    for i in node_indices:
        for j in node_indices:
            if abs(i - j) <= window_size and (i != j or do_self_loops):
                node_adjacency[i][j] = 1

    return node_adjacency


def edge_indices_from_adjacency(node_adjacency: List[List[int]]
                                ) -> List[List[Tuple[int, int]]]:
    node_indices = list(range(len(node_adjacency)))
    return [[i, j] for i in node_indices for j in node_indices if node_adjacency[i][j]]


def graph_dict_to_list_values(g: dict) -> dict:
    """
    Given a GraphDict ``g``, this function will make sure that all of the values of that dict are converted
    to lists. So usually the values of a GraphDict are numpy arrays, but we also want to be able to serialize
    these data structures as JSON strings and that is not possible with numpy arrays, instead we need the
    native representation as nested lists for this.

    :param GraphDict g: The graph representation to be converted

    :returns GraphDict: All of the values of this dict will be nested lists instead of numpy arrays.
    """
    result = {}
    for key, value in g.items():
        if isinstance(value, np.ndarray):
            result[key] = value.tolist()
        else:
            result[key] = value

    return result


def graph_dict_to_numpy_values(g: dict) -> dict:
    """
    Given a GraphDict ``g``, this function will make sure that all the values are converted to numpy arrays.
    When loading graph structures from JSON files, the values are still represented as nested lists, but for
    various operations on GraphDicts, the values need to be numpy arrays and this function makes sure all
    the values are properly converted.

    :param GraphDict g: The graph representation to be converted

    :returns GraphDict: All value of this dict will be numpy arrays instead of nested lists.
    """
    result = {}
    for key, value in g.items():
        result[key] = np.array(value)

    return result


def importance_absolute_similarity(importances_true: list,
                                   importances_pred: list,
                                   normalize: bool = True
                                   ) -> float:
    importances_true = np.array(importances_true)
    importances_pred = np.array(importances_pred)

    if normalize:
        norm = mcolors.Normalize(vmin=0, vmax=np.max(importances_pred))
        importances_pred = np.vectorize(norm)(importances_pred)

    return 1 - np.mean(np.abs(importances_true - importances_pred))


def mask_distance(y_true: np.ndarray, y_pred: np.ndarray, cutoff=0.1) -> float:
    # return distance.canberra(y_true + epsilon, y_pred + epsilon) / len(y_true)
    y_pred_mod = np.copy(y_pred)
    y_pred_mod[y_pred < cutoff] = 0
    return distance.canberra(y_true, y_pred_mod) / len(y_true)

    y_true_inv = 1 - y_true
    y_pred_inv = 1 - y_pred
    return distance.cosine(y_true_inv, y_pred_inv)


def importance_canberra_similarity(importances_true: list,
                                   importances_pred: list,
                                   normalize: bool = True
                                   ) -> Optional[float]:
    # Make sure that we are working with numpy arrays
    importances_true = np.array(importances_true)
    importances_pred = np.array(importances_pred)

    if normalize:
        norm = mcolors.Normalize(vmin=0, vmax=np.max(importances_pred))
        importances_pred = np.vectorize(norm)(importances_pred)

    if len(importances_pred.shape) == 1:
        dist = mask_distance(importances_true, importances_pred)

    if len(importances_pred.shape) == 2:
        distances = [mask_distance(importances_true[:, k], importances_pred[:, k])
                     for k in range(importances_pred.shape[1])]
        dist = np.mean(distances)

    return 1 - dist


def array_normalize(array: np.ndarray
                    ) -> np.ndarray:
    norm = mcolors.Normalize(vmin=np.min(array), vmax=np.max(array))
    return np.vectorize(norm)(array)


def binary_threshold(array: np.ndarray,
                     threshold: float,
                     ) -> np.ndarray:
    binary = np.zeros_like(array)
    binary[array > threshold] = 1
    return binary


# https://stackoverflow.com/questions/6910641/how-do-i-get-indices-of-n-maximum-values-in-a-numpy-array
def binary_threshold_k(array: np.array,
                       k: int,
                       inverse: bool = False
                       ) -> np.array:
    indices = np.argpartition(array, -k)[-k:]
    if inverse:
        binary = np.ones_like(array)
        binary[indices] = 0
    else:
        binary = np.zeros_like(array)
        binary[indices] = 1

    return binary

# == NATURAL LANGUAGE PROCESSING ============================================================================


def load_glove_map(path: str,
                   line_buffer_size: int = 10
                   ) -> Dict[str, list]:
    """
    Given the ``path`` to a file of pre-trained transformations of a GLOVE word to vector model, this
    function will load that file into memory as a defaultdict. The keys of this dict are the string tokens
    and the values are lists with the corresponding numeric vectors. The default case will return a
    zeros array with the same dimensions.

    :param str path: The absolute path to the glove txt file to be used
    :param int line_buffer_size: The number of lines to read from the file at a time. The reading is done
        in a buffered manner because the glove files can get extremely large.
    :return: A dict which maps string tokens to lists of numbers
    """
    glove_map = defaultdict(list)

    num_dimensions = None
    line_index = 0
    with open(path, mode='r') as file:

        lines = file.readlines(line_buffer_size)
        while lines:

            for line in lines:
                token, *values = line.split(' ')
                values = [float(value) for value in values]
                glove_map[token.lower()] = values

                if num_dimensions is None:
                    num_dimensions = len(values)
                elif num_dimensions != len(values):
                    raise ValueError(f'Line {line_index} has a different dimension than previously: '
                                     f'{num_dimensions} != {len(values)}')

                line_index += 1

            lines = file.readlines(line_buffer_size)

    glove_map.default_factory = lambda: [0] * num_dimensions

    return glove_map


def token_dict_to_graph(token_dict: Dict[int, Any],
                        window_size: int = 2,
                        do_self_loops: bool = False
                        ) -> dict:
    g = defaultdict(list)
    for node_index, (token_index, token_attributes) in enumerate(token_dict.items()):
        g['node_indices'].append(node_index)
        g['token_indices'].append(token_index)
        g['node_attributes'].append(token_attributes)

    # Now that we basically already have all of the nodes, we need to connect them up with edges
    # according to the specified window size.
    g['node_adjacency'] = node_adjacency_sliding_window(
        node_indices=g['node_indices'],
        window_size=window_size,
        do_self_loops=do_self_loops
    )

    # Based on the adjacency matrix we can then create the list of edge index tuples and the edge attributes
    # weights, which are just all constant ones in this case
    for i in g['node_indices']:
        for j in g['node_indices']:
            if g['node_adjacency'][i][j]:
                g['edge_indices'].append([i, j])
                g['edge_attributes'].append([1.])

    return dict(g)


graph_from_token_dict = token_dict_to_graph


def text_to_graph(text: str,
                  glove_map: Dict[str, list],
                  sanitize_tokens: list = SANITIZE_TOKENS,
                  window_size: int = 2,
                  do_self_loops: bool = False,
                  include_node_strings: bool = False,
                  ) -> Dict[int, np.ndarray]:
    # ~ 1. Convert string to token dict
    token_list: List[str] = word_tokenize(text)
    token_dict: Dict[int, str] = {i: t.lower()
                                  for i, t in enumerate(token_list)
                                  if t.lower() not in sanitize_tokens}

    # ~ 2. Convert string tokens to vectors with GLOVE
    embedding_dict: Dict[int, np.ndarray] = {i: glove_map[t] for i, t in token_dict.items()}
    g = graph_from_token_dict(
        embedding_dict,
        window_size=window_size,
        do_self_loops=do_self_loops
    )

    if include_node_strings:
        g['node_strings'] = list(token_dict.values())

    # ~ 3. Convert graph properties to np arrays
    g = {k: np.array(v) for k, v in g.items()}

    return g


def latex_table_element_mean(values: List[float],
                             template_name: str = 'table_element_mean.tex.j2',
                             vertical: bool = True,
                             raw: bool = False) -> str:
    if raw:
        mean, std = values
    else:
        mean = np.mean(values)
        std = np.std(values)

    template = TEMPLATE_ENV.get_template(template_name)
    return template.render(
        mean=mean,
        std=std,
        vertical=vertical
    )


def latex_table_element_median(values: List[float],
                               upper_quantile: float = 0.75,
                               lower_quantile: float = 0.25,
                               include_variance: bool = True,
                               template_name: str = 'table_element_median.tex.j2') -> str:
    median = np.median(values)
    upper = np.quantile(values, upper_quantile)
    lower = np.quantile(values, lower_quantile)

    template = TEMPLATE_ENV.get_template(template_name)
    return template.render(
        median=median,
        upper=upper,
        lower=lower,
        include_variance=include_variance
    )


def latex_table(column_names: List[str],
                rows: List[Union[List[float], str]],
                content_template_name: str = 'table_content.tex.j2',
                table_template_name: str = 'table.tex.j2',
                list_element_cb: Callable[[List[float]], str] = latex_table_element_mean,
                prefix_lines: List[str] = [],
                caption: str = '',
                ) -> Tuple[str, str]:

    # ~ Pre Processing the row elements into strings
    string_rows = []
    for row_index, row in enumerate(rows):
        string_row = []
        for element in row:
            if isinstance(element, str):
                string_row.append(element)
            if isinstance(element, list) or isinstance(element, np.ndarray):
                string = list_element_cb(element)
                string_row.append(string)

        string_rows.append(string_row)

    alignment = ''.join(['c' for _ in column_names])

    # ~ Rendering the latex template(s)

    content_template = TEMPLATE_ENV.get_template(content_template_name)
    content = content_template.render(rows=string_rows)

    table_template = TEMPLATE_ENV.get_template(table_template_name)
    table = table_template.render(
        alignment=alignment,
        column_names=column_names,
        content=content,
        header='\n'.join(prefix_lines),
        caption=caption,
    )

    return content, table


def render_latex(kwargs: dict,
                 output_path: str,
                 template_name: str = 'article.tex.j2'
                 ) -> None:
    with tempfile.TemporaryDirectory() as temp_path:
        # First of all we need to create the latex file on which we can then later invoke "pdflatex"
        template = TEMPLATE_ENV.get_template(template_name)
        latex_string = template.render(**kwargs)
        latex_file_path = os.path.join(temp_path, 'main.tex')
        with open(latex_file_path, mode='w') as file:
            file.write(latex_string)

        # Now we invoke the system "pdflatex" command
        command = (f'pdflatex  '
                   f'-interaction=nonstopmode '
                   f'-output-format=pdf '
                   f'-output-directory={temp_path} '
                   f'{latex_file_path} ')
        proc = subprocess.run(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if proc.returncode != 0:
            raise ChildProcessError(f'pdflatex command failed! Maybe pdflatex is not properly installed on '
                                    f'the system? Error: {proc.stdout.decode()}')

        # Now finally we copy the pdf file - currently in the temp folder - to the final destination
        pdf_file_path = os.path.join(temp_path, 'main.pdf')
        shutil.copy(pdf_file_path, output_path)


# == CUSTOM EXCEPTIONS ======================================================================================


class DatasetError(Exception):
    pass




import os
import sys
import click

from pycomex.experiment import run_experiment
from pycomex.cli import ExperimentCLI

from graph_attention_student.util import get_version
from graph_attention_student.util import DATASETS_FOLDER
from graph_attention_student.util import EXAMPLES_FOLDER
from graph_attention_student.util import EXPERIMENTS_PATH

EXAMPLES = [name.split('.')[0] for name in os.listdir(EXAMPLES_FOLDER)]


cli = ExperimentCLI(name='megan', experiments_path=EXPERIMENTS_PATH, version=get_version())


@click.command(short_help='executes the example code')
@click.option('-n', '--name', type=click.Choice(EXAMPLES), default='solubility_regression',
              help='the name of the example file to be executed')
@click.pass_context
def example(ctx, name: str):
    """
    Executes the example code
    """
    click.secho(f'Running example code: {name}!')
    example_path = os.path.join(EXAMPLES_FOLDER, name + '.py')
    experiment_path, proc = run_experiment(example_path)

    click.secho(f'Experiment completed and saved to: {experiment_path}')


cli.add_command(example)

if __name__ == '__main__':
    cli()



"""
Trains multi-channel MEGAN models and a GNNExplainer baseline on the synthetic graph regression dataset
RbMotifs. This datasets consits of randomly generated colored graphs, where the target value of each graph
is determined by the special subgraph motifs contained in it. The MEGAN models are configured to the special
single-explanation channel case, where only a single attention explanation mask is generated for each
sample. The generated explanations are compared to the known ground truth explanations. In the GNNExplainer
case a simple GCN network is trained on the same dataset and for each sample of the test set a new
GNNExplainer optimization loop is performed to generate the explanations. The experiment is repeated
multiple independent times.
"""
import os
import warnings
warnings.filterwarnings("ignore")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import sys
import json
import time
import random
from typing import List, Callable, Dict, Optional
from pprint import pprint
from collections import Counter

import tensorflow as tf
import tensorflow.keras as ks
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from matplotlib.backends.backend_pdf import PdfPages
from sklearn.metrics import roc_auc_score, r2_score
from scipy.stats import wilcoxon
from imageio.v2 import imread
from kgcnn.literature.GNNExplain import GNNExplainer
from kgcnn.utils.data import ragged_tensor_from_nested_numpy

from pycomex.experiment import Experiment
from pycomex.util import Skippable

from graph_attention_student.util import DATASETS_FOLDER
from graph_attention_student.util import importance_absolute_similarity
from graph_attention_student.util import importance_canberra_similarity
from graph_attention_student.util import array_normalize
from graph_attention_student.util import binary_threshold
from graph_attention_student.util import render_latex
from graph_attention_student.util import latex_table, latex_table_element_median
from graph_attention_student.data import load_eye_tracking_dataset
from graph_attention_student.data import process_graph_dataset
from graph_attention_student.models import MultiAttentionStudent
from graph_attention_student.models import GnnxGCN
from graph_attention_student.training import NoLoss, ExplanationLoss
from graph_attention_student.training import LogProgressCallback
from graph_attention_student.visualization import plot_node_importances, plot_edge_importances

SHORT_DESCRIPTION = (
    'Trains single-channel MEGAN model, as well as GNNExplainer, on the synthetic graph regression dataset '
    'RbMotifs and compares the generated explanations to the known ground truth.'
)

# == META PARAMETERS ==
REPETITIONS = 50

# == DATASET PARAMETERS ==
DATASET_PATH = os.path.join(DATASETS_FOLDER, 'rb_dual_motifs')
TEST_RATIO = 0.1
NUM_EXAMPLES = 100
EXAMPLE_INDICES: Optional[list] = None

# == MODEL PARAMETERS ==
UNITS = [9, 9, 9]
IMPORTANCE_UNITS = []
SPARSITY_FACTOR = 0
IMPORTANCE_FACTOR = 0.0
IMPORTANCE_MULTIPLIER = 1
NUM_CHANNELS = 1
FINAL_UNITS = [3]
DROPOUT_RATE = 0.0
FINAL_DROPOUT_RATE = 0.0
USE_BIAS = True
REGRESSION_LIMITS = [-4, 4]
REGRESSION_BINS = [[-4, 0], [0, 4]]
MEGAN_SWEEP = {
    'megan_0': {
        'IMPORTANCE_FACTOR': 0.0,
        'SPARSITY_FACTOR': 0e-2,
        'IMPORTANCE_SUPERVISION': False,
    },
    'megan_1': {
        'IMPORTANCE_FACTOR': 0.0,
        'SPARSITY_FACTOR': 0e-2,
        'IMPORTANCE_SUPERVISION': True
    }
}

# == TRAINING PARAMETERS ==
LEARNING_RATE = 0.004
BATCH_SIZE = 512
EPOCHS = 250
LOG_STEP = 10
DEVICE = 'gpu:0'
METRIC_KEY = 'mean_squared_error'
IMPORTANCE_SUPERVISION = False

# == GNN EXPLAINER PARAMETERS ==
GNNX_MODEL_UNITS = [32, 32, 32]
GNNX_EPOCHS = 100
GNNX_LEARNING_RATE = 0.2
GNNX_NODE_WEIGHT = 0.01
GNNX_EDGE_WEIGHT = 0.01

# == EVALUATION PARAMETERS ==
FIDELITY_SPARSITY = 0.3
BINARY_THRESHOLD = 0.5
LOG_STEP_EVAL = 10
NUM_RANDOM_MASKS = 3

# == EXPERIMENT PARAMETERS ==
NAMESPACE = 'rb_motifs_single'
BASE_PATH = os.getcwd()
DEBUG = True

# == HELPER METHODS ==


with Skippable(), (e := Experiment(namespace=NAMESPACE, base_path=BASE_PATH, glob=globals())):
    e.info('Loading the dataset...')
    start_time = time.time()
    eye_tracking_dataset = load_eye_tracking_dataset(DATASET_PATH)
    dataset_size = len(eye_tracking_dataset)
    end_time = time.time()
    e.info(f'Loaded dataset with {dataset_size} elements in {end_time - start_time:.2f} seconds')
    e.status()

    e['metric_keys'] = []
    for rep in range(REPETITIONS):
        e.info(f'STARTING REPETITION ({rep+1}/{REPETITIONS})')

        dataset_indices = list(range(dataset_size))
        test_indices = random.sample(dataset_indices, k=int(TEST_RATIO * dataset_size))
        train_indices = [index for index in dataset_indices if index not in test_indices]
        e[f'test_indices/{rep}'] = test_indices
        e[f'train_indices/{rep}'] = train_indices
        e.info(f'Identified {len(train_indices)} elements for training and {len(test_indices)} for testing')

        example_indices = random.sample(test_indices, k=NUM_EXAMPLES)
        e[f'example_indices/{rep}'] = example_indices
        e.info(f'Identified {len(example_indices)} elements as examples')

        # "eye_tracking_dataset" at this point is only a list of dictionaries which contain the metadata info
        # for each respective elements. This now has to be turned into the appropriate tensors to be used by
        # the model.
        dataset = []
        for data in eye_tracking_dataset:
            g = data['metadata']['graph']

            # The target ground truth will be the float value associated with every graph
            value = data['metadata']['value']
            g['graph_labels'] = np.array(value)

            # Since this is the single-channel case we also need to modify the ground truth node and edge
            # importance annotations here: Currently they are split into two channels, but we need to combine
            # them into a single channel
            node_importances = np.array(g['multi_node_importances'])
            node_importances = np.sum(node_importances, axis=-1, keepdims=True)
            g['node_importances'] = node_importances

            edge_importances = np.array(g['multi_edge_importances'])
            edge_importances = np.sum(edge_importances, axis=-1, keepdims=True)
            g['edge_importances'] = edge_importances

            dataset.append(g)

        x_train, y_train, x_test, y_test = process_graph_dataset(dataset, test_indices)
        e.info('Pre-processed the dataset. Now ready for training')

        # ---------------------------------------------------------------------------------------------------
        # -- MULTI ATTENTION STUDENT ------------------------------------------------------------------------
        # ---------------------------------------------------------------------------------------------------
        for model_name, parameters in MEGAN_SWEEP.items():
            e.info(f'Model sweep: {model_name}')
            for parameter_key, parameter_value in parameters.items():
                globals()[parameter_key] = parameter_value
                e.info(f' * {parameter_key}: {eval(parameter_key)}')

            e.info('Training Multi Explanation Graph Attention Network (MEGAN) ...')
            ks.backend.clear_session()
            model: ks.models.Model = MultiAttentionStudent(
                units=UNITS,
                dropout_rate=DROPOUT_RATE,
                importance_factor=IMPORTANCE_FACTOR,
                importance_multiplier=IMPORTANCE_MULTIPLIER,
                importance_units=IMPORTANCE_UNITS,
                importance_channels=NUM_CHANNELS,
                use_bias=USE_BIAS,
                final_units=FINAL_UNITS,
                final_dropout_rate=FINAL_DROPOUT_RATE,
                regression_limits=REGRESSION_LIMITS,
                regression_bins=REGRESSION_BINS,
                sparsity_factor=SPARSITY_FACTOR
            )
            model.compile(
                loss=[
                    ks.losses.MeanSquaredError(),
                    NoLoss() if not IMPORTANCE_SUPERVISION else ExplanationLoss(),
                    NoLoss() if not IMPORTANCE_SUPERVISION else ExplanationLoss(),
                ],
                loss_weights=[
                    1,
                    0 if not IMPORTANCE_SUPERVISION else 1,
                    0 if not IMPORTANCE_SUPERVISION else 1,
                ],
                metrics=[
                    ks.metrics.MeanSquaredError(),
                    ks.metrics.MeanAbsoluteError(),
                ],
                optimizer=ks.optimizers.Adam(learning_rate=LEARNING_RATE),
                run_eagerly=False
            )
            with tf.device(DEVICE):
                history = model.fit(
                    x_train,
                    y_train,
                    batch_size=BATCH_SIZE,
                    epochs=EPOCHS,
                    validation_data=(x_test, y_test),
                    validation_freq=1,
                    callbacks=LogProgressCallback(
                        logger=e.logger,
                        epoch_step=LOG_STEP,
                        identifier=f'val_output_1_{METRIC_KEY}'
                    ),
                    verbose=0
                )
            e.info(f'model parameters: {model.count_params()}')
            e[f'histories/{model_name}/{rep}'] = history.history
            e[f'epochs/{model_name}/{rep}'] = list(range(EPOCHS))

            ni_true = []
            ei_true = []

            ni_pred = []
            ei_pred = []

            # ~ Making all the test set predictions and saving them
            e.info('evaluating test set...')
            for c, index in enumerate(test_indices):
                data = eye_tracking_dataset[index]
                g = dataset[index]

                # First of all we need to query the model to make the actual prediction for the current element
                out_meg, ni_meg, ei_meg = model.predict_single((
                    g['node_attributes'],
                    g['edge_attributes'],
                    g['edge_indices'],
                ))
                e[f'out/true/{rep}/{index}'] = g['graph_labels']
                e[f'out/{model_name}/{rep}/{index}'] = out_meg
                ni_true += np.array(g['node_importances']).flatten().tolist()
                ei_true += np.array(g['edge_importances']).flatten().tolist()

                ni_pred += np.array(ni_meg).flatten().tolist()
                ei_pred += np.array(ei_meg).flatten().tolist()
                # The first transformation we apply to the importances vectors is to apply a normalization on
                # them such that they use the entire value range [0, 1]
                ni_meg = array_normalize(ni_meg)
                ei_meg = array_normalize(ei_meg)
                e[f'ni/{model_name}/{rep}/{index}'] = ni_meg
                e[f'ei/{model_name}/{rep}/{index}'] = ei_meg
                # Then with these relatively raw importances we calculate the cont. similarity value
                # (basically something like a distance metric with a normalized & inverted scale)
                node_sim_meg = importance_absolute_similarity(g['node_importances'], ni_meg)
                edge_sim_meg = importance_absolute_similarity(g['edge_importances'], ei_meg)
                e[f'node_sim/{model_name}/{rep}/{index}'] = node_sim_meg
                e[f'edge_sim/{model_name}/{rep}/{index}'] = edge_sim_meg
                node_sim_can_meg = importance_canberra_similarity(g['node_importances'], ni_meg)
                edge_sim_can_meg = importance_canberra_similarity(g['edge_importances'], ei_meg)
                e[f'node_sim_can/{model_name}/{rep}/{index}'] = node_sim_can_meg
                e[f'edge_sim_can/{model_name}/{rep}/{index}'] = edge_sim_can_meg
                # Then we calculate the natural sparsity percentage of the importances. We do this by first
                # applying a simple threshold to turn the soft importance masks into hard importance masks
                # and then the sparsity is simply the mean of the hard mask.
                sparsity_ni = np.mean(binary_threshold(ni_meg, BINARY_THRESHOLD))
                sparsity_ei = np.mean(binary_threshold(ei_meg, BINARY_THRESHOLD))
                e[f'node_sparsity/{model_name}/{rep}/{index}'] = sparsity_ni
                e[f'edge_sparsity/{model_name}/{rep}/{index}'] = sparsity_ei
                e[f'sparsity_ni/{model_name}/{rep}/{index}'] = sparsity_ni
                e[f'sparsity_ei/{model_name}/{rep}/{index}'] = sparsity_ei
                # Binary threshold
                node_mask = binary_threshold(ni_meg, BINARY_THRESHOLD)
                edge_mask = binary_threshold(ei_meg, BINARY_THRESHOLD)

                out_m_meg, _, _ = model.predict_single((
                    np.array(g['node_attributes']) * (1 - node_mask),
                    np.array(g['edge_attributes']) ,#* edge_mask,
                    g['edge_indices']
                ))
                fidelity_meg = abs(out_m_meg - out_meg)
                e[f'fidelity/{model_name}/{rep}/{index}'] = fidelity_meg

                # In the last step we generate a random explanation and then calculate the auroc and the
                # fidelity for that one which will act as a baseline to understand the other values better

                deltas_rand = []
                for _ in range(NUM_RANDOM_MASKS):
                    node_mask_rand_indices = random.sample(g['node_indices'], k=int(np.sum(node_mask)))
                    node_mask_rand = np.array([0.0 if i in node_mask_rand_indices else 1.0
                                               for i in g['node_indices']])
                    node_mask_rand = np.expand_dims(node_mask_rand, axis=-1)
                    e[f'ni_rand/{model_name}/{rep}/{index}'] = node_mask_rand

                    # edge_mask_rand_indices = random.sample(g['edge_indices'], k=edge_mask_count)
                    # edge_mask_rand = np.array([0.0 if m in edge_mask_rand_indices else 1.0
                    #                            for m in g['edge_indices']])
                    # edge_mask_rand = np.expand_dims(edge_mask_rand, axis=-1)
                    # e[f'ei_rand/{model_name}/{rep}/{index}'] = edge_mask_rand

                    out_rand_meg, _, _ = model.predict_single((
                        np.array(g['node_attributes']) * node_mask_rand,
                        np.array(g['edge_attributes']) ,#* edge_mask_rand,
                        g['edge_indices']
                    ))
                    deltas_rand.append(abs(out_rand_meg - out_meg))
                fidelity_rand_meg = np.mean(deltas_rand)
                e[f'fidelity_rand/{model_name}/{rep}/{index}'] = fidelity_rand_meg

                if c % LOG_STEP_EVAL == 0:
                    e.info(f' * MAS ({c}) '
                           f' - node sim can: {node_sim_can_meg:.2f}'
                           f' - edge sim can: {edge_sim_can_meg:.2f}'
                           f' - fidelity: {fidelity_meg:.2f}'
                           f' - fidelity_rand: {fidelity_rand_meg:.2f}'
                           f' - y_true: {e[f"out/true/{rep}/{index}"]:.2f}')

            fidelity_mean = np.mean(list(e[f'fidelity/{model_name}/{rep}'].values()))
            e[f'fidelity_mean/{model_name}/{rep}'] = fidelity_mean
            fidelity_rand_mean = np.mean(list(e[f'fidelity_rand/{model_name}/{rep}'].values()))
            e[f'fidelity_rand_mean/{model_name}/{rep}'] = fidelity_rand_mean

            wilcoxon_result = wilcoxon(
                list(e[f'fidelity/{model_name}/{rep}'].values()),
                list(e[f'fidelity_rand/{model_name}/{rep}'].values())
            )
            e[f'fidelity_wilcoxon/{model_name}/{rep}'] = wilcoxon_result.pvalue

            auroc_ni = roc_auc_score(ni_true, ni_pred)
            e[f'auroc_ni/{model_name}/{rep}'] = auroc_ni

            auroc_ei = roc_auc_score(ei_true, ei_pred)
            e[f'auroc_ei/{model_name}/{rep}'] = auroc_ei

            mse = history.history['val_output_1_mean_squared_error'][-1]
            rmse = np.sqrt(mse)
            e[f'mse/{model_name}/{rep}'] = mse
            e[f'rmse/{model_name}/{rep}'] = rmse

            r2 = r2_score(
                list(e[f'out/true/{rep}'].values()),
                list(e[f'out/{model_name}/{rep}'].values())
            )
            e[f'r2/{model_name}/{rep}'] = r2
            e.info(f'node auroc: {auroc_ni:.2f}'
                   f' - edge auroc: {auroc_ei:.2f}'
                   f' - r2: {r2:.2f}'
                   f' - fidelity: {fidelity_mean:.2f}'
                   f' - fidelity_rand: {fidelity_rand_mean:.2f}')

        # ---------------------------------------------------------------------------------------------------
        # -- GNN EXPLAINER ----------------------------------------------------------------------------------
        # ---------------------------------------------------------------------------------------------------
        e.info('Training GCN for gnn explainer...')
        ks.backend.clear_session()
        gnnx_model = GnnxGCN(
            units=GNNX_MODEL_UNITS,
            activation='kgcnn>leaky_relu',
            final_activation='linear',
            final_units=FINAL_UNITS
        )
        gnnx_model.compile(
            loss=[ks.losses.MeanSquaredError()],
            loss_weights=[1],
            metrics=[ks.metrics.MeanSquaredError()],
            optimizer=ks.optimizers.Adam(learning_rate=LEARNING_RATE),
            run_eagerly=False
        )
        with tf.device(DEVICE):
            history = gnnx_model.fit(
                x_train,
                y_train,
                batch_size=BATCH_SIZE,
                epochs=EPOCHS,
                validation_data=(x_test, y_test),
                validation_freq=1,
                callbacks=LogProgressCallback(
                    logger=e.logger,
                    epoch_step=LOG_STEP,
                    identifier=f'val_{METRIC_KEY}'
                ),
                verbose=0
            )
        e[f'histories/gcn/{rep}'] = history.history
        e[f'epochs/gcn/{rep}'] = list(range(EPOCHS))
        e.info(f'parameters: {gnnx_model.count_params()}')

        ni_true = []
        ei_true = []

        ni_pred = []
        ei_pred = []

        # ~ Making all the test set predictions and saving them
        e.info('Evaluating test set...')
        for c, index in enumerate(test_indices):
            data = eye_tracking_dataset[index]
            g = dataset[index]

            gnnx = GNNExplainer(
                gnnx_model,
                compile_options={
                    'loss': ks.losses.mean_squared_error,
                    'optimizer': ks.optimizers.Adam(learning_rate=GNNX_LEARNING_RATE),
                },
                gnnexplaineroptimizer_options={
                    'edge_mask_loss_weight': GNNX_EDGE_WEIGHT,
                    'node_mask_loss_weight': GNNX_NODE_WEIGHT,
                    'feature_mask_loss_weight': 0
                },
                fit_options={
                    'epochs': GNNX_EPOCHS,
                    'batch_size': 1,
                    'verbose': 0,
                }
            )
            x = (
                ragged_tensor_from_nested_numpy([g['node_attributes']]),
                ragged_tensor_from_nested_numpy([g['edge_attributes']]),
                ragged_tensor_from_nested_numpy([g['edge_indices']])
            )
            insp = gnnx.explain(x, inspection=True)
            ni_gnnx, ei_gnnx = gnnx.get_explanation()
            out_gcn = gnnx_model.predict(x).numpy().tolist()[0]
            e[f'out/gnnx/{rep}/{index}'] = out_gcn
            # The first transformation we apply to the importances vectors is to apply a normalization on
            # them such that they use the entire value range [0, 1]
            ni_gnnx = array_normalize(ni_gnnx)
            ei_gnnx = array_normalize(ei_gnnx)
            e[f'ni/gnnx/{rep}/{index}'] = ni_gnnx
            e[f'ei/gnnx/{rep}/{index}'] = ei_gnnx
            ni_true += np.array(g['node_importances']).flatten().tolist()
            ei_true += np.array(g['edge_importances']).flatten().tolist()

            ni_pred += np.array(ni_gnnx).flatten().tolist()
            ei_pred += np.array(ei_gnnx).flatten().tolist()
            # Then with these relatively raw importances we calculate the cont. similarity value (basically
            # something like a distance metric with a normalized & inverted scale)
            node_sim_gnnx = importance_absolute_similarity(g['node_importances'], ni_gnnx)
            edge_sim_gnnx = importance_absolute_similarity(g['edge_importances'], ei_gnnx)
            e[f'node_sim/gnnx/{rep}/{index}'] = node_sim_gnnx
            e[f'edge_sim/gnnx/{rep}/{index}'] = edge_sim_gnnx
            node_sim_can_gnnx = importance_canberra_similarity(g['node_importances'], ni_gnnx)
            edge_sim_can_gnnx = importance_canberra_similarity(g['edge_importances'], ei_gnnx)
            e[f'node_sim_can/gnnx/{rep}/{index}'] = node_sim_can_gnnx
            e[f'edge_sim_can/gnnx/{rep}/{index}'] = edge_sim_can_gnnx
            # Then we calculate the natural sparsity percentage of the importances. We do this by first
            # applying a simple threshold to turn the soft importance masks into hard importance masks
            # and then the sparsity is simply the mean of the hard mask.
            sparsity_ni = np.mean(binary_threshold(ni_gnnx, BINARY_THRESHOLD))
            sparsity_ei = np.mean(binary_threshold(ei_gnnx, BINARY_THRESHOLD))
            e[f'node_sparsity/gnnx/{rep}/{index}'] = sparsity_ni
            e[f'edge_sparsity/gnnx/{rep}/{index}'] = sparsity_ei
            e[f'sparsity_ni/gnnx/{rep}/{index}'] = sparsity_ni
            e[f'sparsity_ei/gnnx/{rep}/{index}'] = sparsity_ei
            # to binary mask
            node_mask = binary_threshold(ni_gnnx, BINARY_THRESHOLD)
            edge_mask = binary_threshold(ei_gnnx, BINARY_THRESHOLD)

            out_m_gcn = gnnx_model.predict([
                x[0] * ragged_tensor_from_nested_numpy([node_mask]),
                x[1] * ragged_tensor_from_nested_numpy([edge_mask]),
                x[2]
            ]).numpy().tolist()[0]
            fidelity_gnnx = abs(out_m_gcn - out_gcn)
            e[f'fidelity/gnnx/{rep}/{index}'] = fidelity_gnnx

            # In the last step we generate a random explanation and then calculate the auroc and the
            # fidelity for that one which will act as a baseline to understand the other values better

            deltas_rand = []
            for _ in range(NUM_RANDOM_MASKS):
                node_mask_rand_indices = random.sample(g['node_indices'], k=int(np.sum(node_mask)))
                node_mask_rand = np.array([0.0 if i in node_mask_rand_indices else 1.0
                                           for i in g['node_indices']])
                node_mask_rand = np.expand_dims(node_mask_rand, axis=-1)
                e[f'ni_rand/{model_name}/{rep}/{index}'] = node_mask_rand

                # edge_mask_rand_indices = random.sample(g['edge_indices'], k=edge_mask_count)
                # edge_mask_rand = np.array([0.0 if m in edge_mask_rand_indices else 1.0
                #                            for m in g['edge_indices']])
                # edge_mask_rand = np.expand_dims(edge_mask_rand, axis=-1)
                # e[f'ei_rand/{model_name}/{rep}/{index}'] = edge_mask_rand

                out_rand_gcn = gnnx_model.predict([
                    x[0] * ragged_tensor_from_nested_numpy([node_mask_rand]),
                    x[1] ,#* ragged_tensor_from_nested_numpy([edge_mask_rand]),
                    x[2]
                ]).numpy().tolist()[0]
                deltas_rand.append(abs(out_rand_gcn - out_gcn))
            fidelity_rand_gnnx = np.mean(deltas_rand)
            e[f'fidelity_rand/gnnx/{rep}/{index}'] = fidelity_rand_gnnx

            if c % LOG_STEP_EVAL == 0:
                e.info(f' § GNNX ({c})'
                       f' - node sim can: {node_sim_can_gnnx:.2f}'
                       f' - edge sim can: {edge_sim_can_gnnx:.2f}'
                       f' - fidelity: {fidelity_gnnx:.2f}({out_gcn:.2f}>{out_m_gcn:.2f})'
                       f' - fidelity_rand: {fidelity_rand_gnnx:.2f}'
                       f' - gnnx_epochs: {len(insp["total_loss"])}'
                       f' - gnnx_loss: {insp["total_loss"][0]:.2f}>{insp["total_loss"][-1]:.2f}')

        fidelity_mean = np.mean(list(e[f'fidelity/gnnx/{rep}'].values()))
        e[f'fidelity_mean/gnnx/{rep}'] = fidelity_mean
        fidelity_rand_mean = np.mean(list(e[f'fidelity_rand/gnnx/{rep}'].values()))
        e[f'fidelity_rand_mean/gnnx/{rep}'] = fidelity_rand_mean

        wilcoxon_result = wilcoxon(
            list(e[f'fidelity/gnnx/{rep}'].values()),
            list(e[f'fidelity_rand/gnnx/{rep}'].values())
        )
        e[f'fidelity_wilcoxon/gnnx/{rep}'] = wilcoxon_result.pvalue

        auroc_ni = roc_auc_score(ni_true, ni_pred)
        e[f'auroc_ni/gnnx/{rep}'] = auroc_ni

        auroc_ei = roc_auc_score(ei_true, ei_pred)
        e[f'auroc_ei/gnnx/{rep}'] = auroc_ei

        mse = history.history['val_mean_squared_error'][-1]
        rmse = np.sqrt(mse)
        e[f'mse/gnnx/{rep}'] = mse
        e[f'rmse/gnnx/{rep}'] = rmse

        r2 = r2_score(
            list(e[f'out/true/{rep}'].values()),
            list(e[f'out/gnnx/{rep}'].values())
        )
        e[f'r2/gnnx/{rep}'] = r2
        e.info(f'node auroc: {auroc_ni:.2f}'
               f' - edge auroc: {auroc_ei:.2f}'
               f' - r2: {r2:.2f}'
               f' - fidelity: {fidelity_mean:.2f}'
               f' - fidelity_rand: {fidelity_rand_mean:.2f}')

        # ~ Drawing the examples
        e.info('Drawing examples...')
        examples_path = os.path.join(e.path, f'{rep}_examples.pdf')
        ncols = 2 + len(MEGAN_SWEEP)
        with PdfPages(examples_path) as pdf:
            for index in example_indices:
                g = dataset[index]
                data = eye_tracking_dataset[index]
                g['node_coordinates'] = np.array(g['node_coordinates'])
                image = np.asarray(imread(data['image_path']))

                fig, rows = plt.subplots(ncols=ncols, nrows=1, figsize=(ncols * 8, 8), squeeze=False)
                fig.suptitle(f'Rep {rep} - Element {index}\n'
                             f'Ground Truth: {e[f"out/true/{rep}/{index}"]:.3f}')

                column_index = 0
                # -- GROUND TRUTH --
                ax_gt = rows[0][column_index]
                ax_gt.set_title('Ground Truth\n\n')
                ax_gt.imshow(image, extent=(0, image.shape[0], 0, image.shape[1]))
                plot_node_importances(
                    g=g,
                    ax=ax_gt,
                    vmax=np.max(g['node_importances']),
                    node_importances=np.squeeze(g['node_importances'], axis=-1),
                    node_coordinates=g['node_coordinates']
                )
                plot_edge_importances(
                    g=g,
                    ax=ax_gt,
                    vmax=np.max(g['edge_importances']),
                    edge_importances=np.squeeze(g['edge_importances'], axis=-1),
                    node_coordinates=g['node_coordinates']
                )
                column_index += 1

                # -- MULTI ATTENTION STUDENT --
                for model_name in MEGAN_SWEEP.keys():
                    ax_mas = rows[0][column_index]
                    ax_mas.set_title(f'Model "{model_name}"\n'
                                     f'Prediction: {e[f"out/{model_name}/{rep}/{index}"]:.2f}\n'
                                     f'ni sim can: {e[f"node_sim_can/{model_name}/{rep}/{index}"]:.2f} - '
                                     f'ei sim can: {e[f"edge_sim_can/{model_name}/{rep}/{index}"]:.2f}\n '
                                     f'fid: {e[f"fidelity/{model_name}/{rep}/{index}"]:.2f} - '
                                     f'fid rand: {e[f"fidelity_rand/{model_name}/{rep}/{index}"]:.2f}')
                    ax_mas.imshow(image, extent=(0, image.shape[0], 0, image.shape[1]))
                    ni_mas = np.array(e[f'ni/{model_name}/{rep}/{index}'])
                    plot_node_importances(
                        g=g,
                        ax=ax_mas,
                        vmax=np.max(ni_mas),
                        node_importances=np.squeeze(ni_mas, axis=-1),
                        node_coordinates=g['node_coordinates']
                    )
                    ei_mas = np.array(e[f'ei/{model_name}/{rep}/{index}'])
                    plot_edge_importances(
                        g=g,
                        ax=ax_mas,
                        vmax=np.max(ei_mas),
                        edge_importances=np.squeeze(ei_mas, axis=-1),
                        node_coordinates=g['node_coordinates']
                    )
                    column_index += 1

                # -- GNN EXPLAINER --
                ax_gnnx = rows[0][column_index]
                model_name = 'gnnx'
                ax_gnnx.set_title(f'Model "{model_name}"\n'
                                  f'Prediction: {e[f"out/{model_name}/{rep}/{index}"]:.2f}\n'
                                  f'ni sim can: {e[f"node_sim_can/{model_name}/{rep}/{index}"]:.2f} - '
                                  f'ei sim can: {e[f"edge_sim_can/{model_name}/{rep}/{index}"]:.2f}\n '
                                  f'fid: {e[f"fidelity/{model_name}/{rep}/{index}"]:.2f} - '
                                  f'fid rand: {e[f"fidelity_rand/{model_name}/{rep}/{index}"]:.2f}')
                ax_gnnx.imshow(image, extent=(0, image.shape[0], 0, image.shape[1]))
                ni_gnnx = np.array(e[f'ni/gnnx/{rep}/{index}'])
                plot_node_importances(
                    g=g,
                    ax=ax_gnnx,
                    vmax=np.max(ni_gnnx),
                    node_importances=np.squeeze(ni_gnnx, axis=-1),
                    node_coordinates=g['node_coordinates']
                )
                ei_gnnx = np.array(e[f'ei/gnnx/{rep}/{index}'])
                plot_edge_importances(
                    g=g,
                    ax=ax_gnnx,
                    vmax=np.max(ei_gnnx),
                    edge_importances=np.squeeze(ei_gnnx, axis=-1),
                    node_coordinates=g['node_coordinates']
                )
                column_index += 1

                pdf.savefig(fig)
                plt.close(fig)

        e.info(f'finished repetition {rep+1}')
        e.status()


with Skippable(), e.analysis:
    model_names = ['gnnx'] + list(MEGAN_SWEEP.keys())
    # This flag determines whether runs where one of the models did not converge should be excluded
    CLEAN_RESULTS = True

    CLEAN_REPETITIONS = list(range(REPETITIONS))
    if CLEAN_RESULTS:
        e.info('excluding non-convergent repetitions...')
        for rep in range(REPETITIONS):
            for model_name in model_names:
                if e[f'mse/{model_name}/{rep}'] > 2.5:
                    CLEAN_REPETITIONS.remove(rep)
                    break

        e.info(f'{len(CLEAN_REPETITIONS)} repetitions remaining')

    # -- POST CALCULATIONS --
    e.info('post hoc calculations...')
    for rep in range(REPETITIONS):
        for model_name in model_names:
            r2 = r2_score(
                list(e[f'out/true/{rep}'].values()),
                list(e[f'out/{model_name}/{rep}'].values())
            )
            e[f'r2/{model_name}/{rep}'] = r2

            e[f'ni_auroc/{model_name}/{rep}'] = e[f'auroc_ni/{model_name}/{rep}']
            e[f'ei_auroc/{model_name}/{rep}'] = e[f'auroc_ei/{model_name}/{rep}']

    # -- SIMPLE STATISTICS --
    e.info('Printing simple statistics')
    REPETITIONS = e.parameters['REPETITIONS']
    metrics = ['node_sim', 'edge_sim', 'node_sim_can', 'edge_sim_can',
               'node_sparsity', 'edge_sparsity',
               'fidelity', 'fidelity_rand']

    for metric in metrics:
        e.info('')
        e.info(metric.upper())
        for model_name in model_names:
            metric_values = []
            for rep in CLEAN_REPETITIONS:
                for index in e[f'{metric}/{model_name}/{rep}'].keys():
                    value = e[f'{metric}/{model_name}/{rep}/{index}']
                    if value is not None:
                        metric_values.append(value)

            e.info(f'')
            e.info(f'* {model_name} / {metric}')
            e.info(f'  - min: {np.min(metric_values):.2f} - max: {np.max(metric_values):.2f}')
            e.info(f'  - mean: {np.mean(metric_values):.2f}: (std: {np.std(metric_values):.2f})')
            e.info(f'  - (0.25 - median - 0.75): ({np.quantile(metric_values, 0.25):.2f} - '
                   f'{np.median(metric_values):.2f} - {np.quantile(metric_values, 0.75):.2f})')

    # -- RENDERING LATEX TABLE --
    e.info('rendering latex table...')
    column_names = [
        'Model',
        'MSE',
        '$R^2$',
        'Node AUROC',
        'Edge AUROC',
        'Sparsity',
        'Fidelity',
        'Fidelity Rand.'
    ]
    rows = []
    for model_name in model_names:
        row = []
        if 'megan' in model_name:
            parameters = MEGAN_SWEEP[model_name]
            model_identifier = r'$\text{MEGAN}^1_{' + f'{parameters["IMPORTANCE_FACTOR"]:.2f}' + r'}$'
        else:
            model_identifier = r'$\text{GNNX}_{\text{GCN}}$'
        row.append(model_identifier)
        row.append([e[f'mse/{model_name}/{rep}']
                    for rep in CLEAN_REPETITIONS])
        row.append([e[f'r2/{model_name}/{rep}']
                    for rep in CLEAN_REPETITIONS])
        row.append([e[f'auroc_ni/{model_name}/{rep}']
                    for rep in CLEAN_REPETITIONS])
        row.append([e[f'auroc_ei/{model_name}/{rep}']
                    for rep in CLEAN_REPETITIONS])
        row.append([e[f'node_sparsity/{model_name}/{rep}/{index}']
                    for rep in CLEAN_REPETITIONS for index in e[f'test_indices/{rep}']])
        row.append([e[f'fidelity/{model_name}/{rep}/{index}']
                    for rep in CLEAN_REPETITIONS for index in e[f'test_indices/{rep}']])
        row.append([e[f'fidelity_rand/{model_name}/{rep}/{index}']
                    for rep in CLEAN_REPETITIONS for index in e[f'test_indices/{rep}']])
        rows.append(row)

    content, table = latex_table(
        column_names=column_names,
        rows=rows,
        list_element_cb=latex_table_element_median
    )
    e.commit_raw('table.tex', table)
    pdf_path = os.path.join(e.path, 'table.pdf')
    render_latex({'content': table}, output_path=pdf_path)
    e.info('rendered latex table')


from collections import defaultdict
from typing import List, Dict, Callable, Optional
import typing as t

import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import matplotlib.patches as mpatches
from matplotlib.backends.backend_pdf import PdfPages
from imageio.v2 import imread

from graph_attention_student.typing import GraphDict, RgbList


reds_cmap: mcolors.Colormap = mcolors.LinearSegmentedColormap.from_list(
    'reds',
    [
        '#FFFFFF',
        '#FF6F00',
        '#8E0000'
    ]
)


# == MISC. VISUALIZATIONS ==

def plot_regression_fit(values_true: t.Union[np.ndarray, t.List[float]],
                        values_pred: t.Union[np.ndarray, t.List[float]],
                        ax: plt.Axes,
                        num: int = 40,
                        cmap: t.Union[str, mcolors.Colormap] = reds_cmap,
                        plot_reference: bool = True,
                        reference_color: str = 'gray',
                        ):
    """
    Plots a regression plot in the Axes ``ax`` using the true target values ``values_true`` and the
    predicted values ``values_pred``. The plot is in the format of a heatmap, where ``num`` defines the
    granularity of that heat map in x and y direction of the plot.

    It is important that the arrays for the true and predicted values have the same number of elements!

    :param values_true: An array containing the ground truth float target values.
    :param values_pred: An array containing the predicted values which were created by some sort of
        regression model.
    :param ax: The matplotlib Axes on which to draw the plot
    :param num: The integer number which defines the granularity of the heatmap plot. The heatmap is
        discretized in x and y direction using binning with ``num`` bins.
    :param cmap: The matplotlib color map, to be used for the heatmap visualization.
    :param plot_reference: This flag determines if the regression reference line is also to be drawn onto
        the plot. The reference line is a line from the bottom left corner to the top right corner of the
        plot. If all the elements of the regression plot are exactly on that line, it is considered to be a
        perfect regression result.
    :param reference_color: The color to be used for the reference line.

    :return: The binned value map which is the basis for the heatmap.
    """
    # 27.03.2023 - Here we turn the input into a numpy array but more important is the squeeze operation,
    # which fixes a bug, where the function cannot be directly applied to the output of a neural network
    # as that has an additional (redundant) dimension for the output tensor.
    values_true = np.squeeze(np.array(values_true))
    values_pred = np.squeeze(np.array(values_pred))

    min_value = min(np.min(values_true), np.min(values_pred))
    max_value = max(np.max(values_true), np.max(values_pred))
    xs = np.linspace(min_value, max_value, num)
    ys = np.linspace(min_value, max_value, num)
    z, x_edges, y_edges = np.histogram2d(
        x=values_true,
        y=values_pred,
        bins=[xs, ys],
    )

    y, x = np.meshgrid(
        np.linspace(min_value, max_value, num),
        np.linspace(min_value, max_value, num)
    )
    ax.set_xlabel('true values')
    ax.set_xlim([min_value, max_value])
    ax.set_ylabel('predicted values')
    ax.set_ylim([min_value, max_value])
    c = ax.pcolormesh(
        x, y, z,
        linewidth=0,
        cmap=cmap,
        rasterized=True,
    )

    if plot_reference:
        ax.plot(
            [min_value, max_value],
            [min_value, max_value],
            color=reference_color,
            zorder=1,
        )

    return c

# == EXPLANATION VISUALIZATION ===

def plot_node_importances(g: dict,
                          node_importances: np.ndarray,
                          node_coordinates: np.ndarray,
                          ax: plt.Axes,
                          radius: float = 30,
                          thickness: float = 4,
                          color='black',
                          vmin: float = 0,
                          vmax: float = 1):
    node_normalize = mcolors.Normalize(vmin=vmin, vmax=vmax)
    for i, (x, y), ni in zip(g['node_indices'], node_coordinates, node_importances):
        circle = plt.Circle(
            (x, y),
            radius=radius,
            lw=thickness,
            color=color,
            fill=False,
            alpha=node_normalize(ni)
        )
        ax.add_artist(circle)


def plot_edge_importances(g: dict,
                          edge_importances: np.ndarray,
                          node_coordinates: np.ndarray,
                          ax: plt.Axes,
                          radius: float = 30,
                          thickness: float = 4,
                          color='black',
                          vmin: float = 0,
                          vmax: float = 1):
    edge_normalize = mcolors.Normalize(vmin=vmin, vmax=vmax)
    for (i, j), ei in zip(g['edge_indices'], edge_importances):
        coords_i = node_coordinates[i]
        coords_j = node_coordinates[j]
        # Here we determine the actual start and end points of the line to draw. Now we cannot simply use
        # the node coordinates, because that would look pretty bad. The lines have to start at the very
        # edge of the node importance circle which we have already drawn (presumably) at this point. This
        # circle is identified by it's radius. So what we do here is we reduce the length of the line on
        # either side of it by exactly this radius. We do this by first calculating the unit vector for that
        # line segment and then moving radius times this vector into the corresponding directions
        diff = (coords_j - coords_i)
        delta = (radius / np.linalg.norm(diff)) * diff
        x_i, y_i = coords_i + delta
        x_j, y_j = coords_j - delta

        ax.plot(
            [x_i, x_j],
            [y_i, y_j],
            color=color,
            lw=thickness,
            alpha=edge_normalize(ei)
        )


def pdf_from_eye_tracking_dataset(eye_tracking_dataset: List[dict],
                                  pdf_path: str,
                                  include_meta: bool = True,
                                  name: str = 'Eye Tracking Dataset'):

    with PdfPages(pdf_path) as pdf:

        is_classification = False
        if include_meta:
            fig, (ax_task, ax_sizes) = plt.subplots(nrows=1, ncols=2, figsize=(16, 8))
            fig.suptitle(name)

            # This plot is meant to show the label distribution for the classification task
            is_classification = eye_tracking_dataset[0]['metadata']['input_type'] == 'classification'
            if is_classification:
                ax_task.set_title("Label Distribution")
                classes = eye_tracking_dataset[0]['metadata']['classes']
                label_counts = defaultdict(int)
                for data in eye_tracking_dataset:
                    label = data['metadata']['label']
                    label_counts[label] += 1
                xs = list(range(len(label_counts)))
                ax_task.bar(xs, label_counts.values(), color='lightgray')
                ax_task.set_xticks(xs)
                ax_task.set_xticklabels([classes[k] for k in label_counts.keys()])

            # This plot is meant to show the size distribution
            ax_sizes.set_title('Graph Size Distribution')
            sizes = [len(data['metadata']['graph']['node_indices']) for data in eye_tracking_dataset]
            hist, bin_edges = np.histogram(sizes, bins=10)
            xs = list(range(len(hist)))
            ax_sizes.bar(xs, hist, color='lightgray')
            ax_sizes.set_xticks(xs)
            ax_sizes.set_xticklabels([f'({int(bin_edges[i])},{int(bin_edges[i+1])})'
                                      for i in range(len(bin_edges) - 1)])

            pdf.savefig(fig)

        for index, data in enumerate(eye_tracking_dataset):
            g = data['metadata']['graph']
            image_path = data['image_path']
            image = imread(image_path)
            array = np.asarray(image)

            fig, (ax_img, ax_exp) = plt.subplots(nrows=1, ncols=2, figsize=(16, 8))
            fig.suptitle(data['metadata']['name'] if 'name' in data['metadata'] else f'Element {index}')
            ax_img.imshow(array, extent=(0, array.shape[0], 0, array.shape[1]))
            if is_classification:
                label = classes[data["metadata"]["label"]]
                ax_img.set_title(f'Target: {label}')

            # The second graph is supposed to show the ground truth importances, if they exist
            ax_exp.set_title(f'Ground Truth Importances')
            ax_exp.imshow(array, extent=(0, array.shape[0], 0, array.shape[1]))
            if 'node_importances' in g:
                plot_node_importances(
                    g=g,
                    node_importances=g['node_importances'],
                    node_coordinates=np.array(g['node_coordinates']),
                    ax=ax_exp
                )
            if 'edge_importances' in g:
                plot_edge_importances(
                    g=g,
                    edge_importances=g['edge_importances'],
                    node_coordinates=np.array(g['node_coordinates']),
                    ax=ax_exp
                )

            pdf.savefig(fig)
            plt.close(fig)


def plot_distribution(ax: plt.Axes,
                      values: t.Union[t.List[float], np.ndarray],
                      bins: int,
                      show_mean: bool = True,
                      show_std: bool = True,
                      line_width: int = 2):
    histogram, bins, patches = ax.hist(values, bins=bins)
    mean = float(np.mean(values))
    std = float(np.std(values))

    # ~ plot the mean
    if show_mean:
        ax.axvline(
            x=mean,
            color='black',
            ls='--',
            linewidth=line_width,
            label=f'mean: {mean:.2f}'
        )

    # ~ Plot the standard deviation
    if show_std:
        y_min, y_max = ax.get_ylim()
        y_middle = np.mean([y_min, y_max])
        ax.plot(
            [mean - std, mean + std],
            [y_middle, y_middle],
            color='black',
            ls='-',
            linewidth=line_width,
            label=f'std: {std:.2f}'
        )

    return {
        'histogram': histogram,
        'bins': bins,
        'patches': patches,
        'mean': mean,
        'std': std
    }


# == COLORED GRAPHS VIS. ====================================================================================

def draw_node_color_scatter(ax: plt.Axes,
                            g: GraphDict,
                            i: int,
                            size: int = 500,
                            alpha: float = 1.0,
                            border_width: float = 2,
                            add_label: bool = False,
                            marker: str = 'o') -> None:
    # First of all we will draw a white circle underneath so that we can safely apply an alpha value for
    # the actual color whithout seeing through the edges potentially
    x, y = g['node_positions'][i]
    ax.scatter(
        x, y,
        s=size,
        color='white',
        marker=marker,
        zorder=0
    )

    rgb_list = g['node_attributes'][i][:3]
    color = mcolors.to_hex(rgb_list)
    ax.scatter(
        x, y,
        s=size,
        color=color,
        marker=marker,
        alpha=alpha,
        linewidths=border_width,
        edgecolors='black',
        zorder=1,
    )


def draw_edge_black(ax: plt.Axes,
                    g: GraphDict,
                    i: int,
                    j: int,
                    size: int = 2) -> None:
    x_i, y_i = g['node_positions'][i]
    x_j, y_j = g['node_positions'][j]

    ax.plot(
        [x_i, x_j],
        [y_i, y_j],
        lw=size,
        color='black',
        zorder=-1,
    )


def draw_extended_colors_graph(ax: plt.Axes,
                               g: GraphDict,
                               layout_cb: Callable[[nx.Graph], List[int]] = nx.layout.kamada_kawai_layout,
                               draw_node_cb: Callable[[plt.Axes, GraphDict, int], None] = draw_node_color_scatter,
                               draw_edge_cb: Callable[[plt.Axes, GraphDict, int, int], None] = draw_edge_black,
                               ):
    if 'node_positions' not in g:
        graph = nx.Graph()
        graph.add_nodes_from({i: {} for i in g['node_indices']})
        graph.add_edges_from(g['edge_indices'])

        pos = layout_cb(graph)
        g['node_positions'] = np.array(list(pos.values()))

    for i in g['node_indices']:
        draw_node_cb(ax, g, i)

    for (i, j) in g['edge_indices']:
        draw_edge_cb(ax, g, i, j)

    return ax, graph


# == NATURAL LANGUAGE PROCESSING (NLP) VIS. =================================================================

def fixed_row_layout(g: dict,
                     ncols: int = 5,
                     column_spacing: float = 0.5,
                     row_spacing: float = 0.3,
                     ) -> List[List[float]]:
    node_positions = []

    x, y = 0, 0
    for i, node_index in enumerate(g['node_indices']):
        node_positions.append([x, y])

        if i % ncols == 0 and i != 0:
            x = 0
            y -= row_spacing
        else:
            x += column_spacing

    return node_positions


def plot_text_graph(g: dict,
                    ax: plt.Axes,
                    node_importances: Optional[List[float]] = None,
                    edge_importances: Optional[List[float]] = None,
                    text_offset: float = 0.05,
                    node_size: float = 40,
                    node_color: str = 'black',
                    edge_color: str = 'black',
                    do_edges: bool = True,
                    vmin: float = 0.0,
                    vmax: float = 1.0,
                    ) -> None:
    colors = [
        [1, 1, 1, 1],
        [0, 1, 0, 1],
    ]
    cmap = mcolors.LinearSegmentedColormap.from_list('', colors, 256)
    norm = mcolors.Normalize(vmin=vmin, vmax=vmax)

    for i in g['node_indices']:
        x, y = g['node_positions'][i]

        ax.scatter(
            x, y,
            color='black',
            zorder=1,
            s=node_size
        )

        if node_importances is not None:
            value = norm(node_importances[i])
            text_color = cmap(value)
        else:
            text_color = 'white'

        ax.text(
            x, y - text_offset,
            s=g['node_strings'][i],
            ha='center',
            color=node_color,
            bbox={
                'boxstyle': 'round',
                'pad': 0.5,
                'ec': 'black',
                'fc': text_color,
            },
            zorder=1
        )

    if do_edges:

        edge_blacklist = set()
        for e, (i, j) in enumerate(g['edge_indices']):
            if (i, j) in edge_blacklist:
                continue

            x_i, y_i = g['node_positions'][i]
            x_j, y_j = g['node_positions'][j]
            edge_alpha = 1.0
            if edge_importances is not None:
                edge_alpha = norm(edge_importances[e])

            word_distance = abs(i - j)
            if word_distance == 1 or y_i != y_j:

                ax.plot(
                    [x_i, x_j],
                    [y_i, y_j],
                    color=edge_color,
                    zorder=-2,
                    alpha=edge_alpha,
                )

            elif word_distance > 1:
                arrow = mpatches.FancyArrowPatch(
                    (x_i, y_i),
                    (x_j, y_j),
                    connectionstyle=f'arc3, rad=-{.25 * word_distance}',
                    color=edge_color,
                    zorder=-2,
                    alpha=edge_alpha,
                )
                ax.add_patch(arrow)

            edge_blacklist.add((i, j))
            edge_blacklist.add((j, i))




"""
This example shows how the internal implementation of GNES can be trained in an explanation supervised
manner using the synthetic rb motifs dataset

**IMPORTANT IMPLEMENTATION CAVEAT**
One important thing to note: Currently we are not able to provide an efficient implementation of GradCAM
(which GNES is based on) using Keras RaggedTensors. In the current implementation this leads to the
restriction that the model needs to know the exact size of each training batch prior to training. This
means that the training dataset has to have a number of elements which is exactly divisible by the
batch size, such that every batch is guaranteed to have the same size!
"""
import os
import warnings
warnings.filterwarnings("ignore")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import random
import typing as t

import tensorflow as tf
import tensorflow.keras as ks
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from pycomex.experiment import Experiment
from pycomex.util import Skippable
from sklearn.metrics import r2_score
from sklearn.metrics import roc_auc_score
from imageio.v2 import imread

import graph_attention_student.typing as tc
from graph_attention_student.util import DATASETS_FOLDER
from graph_attention_student.util import array_normalize
from graph_attention_student.data import process_graph_dataset
from graph_attention_student.data import load_eye_tracking_dataset_dict
from graph_attention_student.training import LogProgressCallback
from graph_attention_student.training import NoLoss, ExplanationLoss
from graph_attention_student.training import mae
from graph_attention_student.models import GnesGradientModel
from graph_attention_student.models import grad_importances, grad_cam_importances
from graph_attention_student.visualization import plot_node_importances, plot_edge_importances


# == DATASET PARAMETERS ==
DATASET_PATH = os.path.join(DATASETS_FOLDER, 'rb_dual_motifs')
METADATA_CONTAINS_INDICES = True
TRAIN_RATIO = 0.9
NUM_EXAMPLES = 50

# == MODEL PARAMETERS ==
MODEL_CLASS = GnesGradientModel
IMPORTANCES_FUNC = lambda *args: grad_importances(*args, use_absolute=True, use_relu=False)
UNITS = [5, 5, 5]

# == TRAINING PARAMETERS ==
EPOCHS = 50
BATCH_SIZE = 64
LEARNING_RATE = 0.01
SPARSITY_FACTOR = 1.0
EXPLANATION_SUPERVISION = True

# == EVALUATION PARAMETERS ==
EVAL_BATCH_SIZE = 300

# == EXPERIMENT PARAMETERS ==
PATH = os.getcwd()
NAMESPACE = 'results/gnes_example'
DEBUG = True
with Skippable(), (e := Experiment(base_path=PATH, namespace=NAMESPACE, glob=globals())):
    e.info('starting gnes example...')

    e.info('loading rb motifs dataset...')
    dataset_dict = load_eye_tracking_dataset_dict(DATASET_PATH, logger=e.logger)
    dataset_size = len(dataset_dict)
    dataset_index_dict = {}
    indices = []
    dataset: t.List[tc.GraphDict] = []
    for c, (name, data) in enumerate(dataset_dict.items()):
        metadata = data['metadata']
        g = metadata['graph']

        if METADATA_CONTAINS_INDICES:
            index = metadata['index']
        else:
            index = c

        g['node_importances'] = np.sum(np.array(g['multi_node_importances']), axis=-1, keepdims=True)
        g['edge_importances'] = np.sum(np.array(g['multi_edge_importances']), axis=-1, keepdims=True)

        g['edge_attributes'] = np.array(g['edge_attributes'], dtype=np.float)

        # With this here we want to support the possibility that the "value" annotation might already
        # be a multi target annotation (as is the case with classification datasets for example) but usually
        # this will be a single regression value and we will have to wrap that into a list first
        if isinstance(metadata['value'], list):
            g['graph_labels'] = np.array(metadata['value'])
        else:
            g['graph_labels'] = np.array([metadata['value']])

        indices.append(index)
        dataset_index_dict[index] = data

    dataset = [t[1]['metadata']['graph'] for t in sorted(dataset_index_dict.items(), key=lambda t: t[0])]

    e.info('determining train test split...')
    # At this point we need to determine the number of training samples exactly such that it is divisible
    # by the batch size! Because the implementation of GradCAM currently requires that
    num_train_samples = int(TRAIN_RATIO * dataset_size)
    if num_train_samples % BATCH_SIZE != 0:
        remaining = num_train_samples % BATCH_SIZE
        num_train_samples -= remaining

    e.info(f'training dataset of size {num_train_samples}')

    train_indices = random.sample(indices, k=num_train_samples)
    test_indices = [index for index in indices if index not in train_indices]
    example_indices = random.sample(test_indices, k=NUM_EXAMPLES)
    x_train, y_train, x_test, y_test = process_graph_dataset(dataset, test_indices=test_indices)

    e.info('creating the model...')
    model: ks.models.Model = MODEL_CLASS(
        units=UNITS,
        num_outputs=1,
        batch_size=BATCH_SIZE,
        importance_func=IMPORTANCES_FUNC,
        sparsity_factor=SPARSITY_FACTOR
    )
    model.compile(
        loss=[
            ks.losses.MeanSquaredError(),
            ExplanationLoss(loss_function=mae) if EXPLANATION_SUPERVISION else NoLoss(),
            ExplanationLoss(loss_function=mae) if EXPLANATION_SUPERVISION else NoLoss(),
        ],
        loss_weights=[
            1,
            1,
            1
        ],
        metrics=[
            ks.metrics.MeanSquaredError(),
            ks.metrics.MeanAbsoluteError(),
        ],
        optimizer=ks.optimizers.Adam(learning_rate=LEARNING_RATE),
        run_eagerly=False
    )

    with tf.device('gpu:0'):
        # NOTE: validation during the training is explicitly disabled because the validation dataset would
        # have to be divisible by the batch size as well. As that would be too much of a hassle we rather
        # disable it and perform the assessment manually afterwards.
        hist = model.fit(
            x_train,
            y_train,
            batch_size=BATCH_SIZE,
            epochs=EPOCHS,
            # validation_data=(x_test, y_test),
            # validation_freq=1,
            callbacks=LogProgressCallback(
                logger=e.logger,
                epoch_step=1,
                identifier=f'output_1_loss'
            ),
            verbose=0
        )
        e.info('finished model training')
        e.info(f'model with {model.count_params()} parameters')
        history = hist.history
        e['history'] = history

        e.info('evaluating on the test set...')
        eval_indices = test_indices.copy()
        current = 0
        while current < len(test_indices):
            num_samples = min(len(test_indices) - current, EVAL_BATCH_SIZE)

            eval_indices = test_indices[current:current+num_samples]
            node_input = x_test[0][current:current+num_samples]
            edge_input = x_test[1][current:current+num_samples]
            edge_index_input = x_test[2][current:current+num_samples]

            x_eval = (node_input, edge_input, edge_index_input)
            out_pred, ni_pred, ei_pred = [v.numpy() for v in model(x_eval, batch_size=num_samples)]
            out_true = y_test[0][current:current+num_samples]
            for c, index in enumerate(eval_indices):
                e[f'out/pred/{index}'] = float(out_pred[c][0])
                e[f'out/true/{index}'] = float(out_true[c][0])
                e[f'ni/{index}'] = array_normalize(ni_pred[c])
                e[f'ei/{index}'] = array_normalize(ei_pred[c])

            current += num_samples
            e.info(f' * evaluated ({current}/{len(test_indices)})')

        e.info('calculating evaluation metrics...')
        errors = []
        ni_true = []
        ni_pred = []
        ei_true = []
        ei_pred = []
        for index in test_indices:
            data = dataset_index_dict[index]
            g = data['metadata']['graph']

            error = e[f'out/true/{index}'] - e[f'out/pred/{index}']
            errors.append(error)

            ni_pred += np.array(e[f'ni/{index}']).flatten().tolist()
            ni_true += np.array(g['node_importances']).flatten().tolist()

            ei_pred += np.array(e[f'ei/{index}']).flatten().tolist()
            ei_true += np.array(g['edge_importances']).flatten().tolist()

        mse = np.mean(np.square(errors))
        mae = np.mean(np.abs(errors))
        e['mse'] = mse
        e['mae'] = mae
        e.info(f' * mse: {mse}')

        r2 = r2_score(list(e['out/true'].values()), list(e['out/pred'].values()))
        e['r2'] = r2
        e.info(f' * r2: {r2}')

        ni_auc = roc_auc_score(ni_true, ni_pred)
        ei_auc = roc_auc_score(ei_true, ei_pred)
        e['ni_auc'] = ni_auc
        e['ei_auc'] = ei_auc
        e.info(f' * ni_auc: {ni_auc}')
        e.info(f' * ei_auc: {ei_auc}')

        e.info('drawing examples...')
        path = os.path.join(e.path, 'examples.pdf')
        num_channels = 1
        with PdfPages(path) as pdf:
            for c, index in enumerate(example_indices):
                ni = np.array(e[f'ni/{index}'])
                ei = np.array(e[f'ei/{index}'])

                data = dataset_index_dict[index]
                metadata = data['metadata']
                g = metadata['graph']
                g['node_coordinates'] = np.array(g['node_coordinates'])
                image = np.asarray(imread(data['image_path']))

                fig, rows = plt.subplots(ncols=2, nrows=num_channels, figsize=(16, 16),
                                         squeeze=False)

                for row_index in range(num_channels):
                    column_index = 0

                    # -- GROUND TRUTH --
                    ax_gt = rows[row_index][column_index]
                    ax_gt.imshow(image, extent=(0, image.shape[0], 0, image.shape[1]))
                    node_importances = np.array(g['node_importances'])
                    plot_node_importances(
                        g=g,
                        ax=ax_gt,
                        vmax=np.max(node_importances[:, row_index]),
                        node_importances=node_importances[:, row_index],
                        node_coordinates=g['node_coordinates']
                    )
                    edge_importances = np.array(g['edge_importances'])
                    plot_edge_importances(
                        g=g,
                        ax=ax_gt,
                        vmax=np.max(edge_importances[:, row_index]),
                        edge_importances=edge_importances[:, row_index],
                        node_coordinates=g['node_coordinates']
                    )
                    column_index += 1

                    # -- EXPLANATIONS --
                    ax_mas = rows[row_index][column_index]
                    ax_mas.imshow(image, extent=(0, image.shape[0], 0, image.shape[1]))
                    plot_node_importances(
                        g=g,
                        ax=ax_mas,
                        vmax=np.max(ni),
                        node_importances=ni[:, row_index],
                        node_coordinates=g['node_coordinates']
                    )
                    plot_edge_importances(
                        g=g,
                        ax=ax_mas,
                        vmax=np.max(ei),
                        edge_importances=ei[:, row_index],
                        node_coordinates=g['node_coordinates']
                    )
                    column_index += 1

                pdf.savefig(fig)
                plt.close(fig)




"""
This file mainly exists to provide the CUSTOM_OBJECTS global dict. This dict will be needed when a model
from this package is to be loaded from a persistent representation, this dictionary has to be passed as
an argument to a keras ``custom_object_scope`` context:

.. code-block:: python

    import tensorflow.keras as ks
    from graph.attention_student.keras import CUSTOM_OBJECTS

    with ks.util.custom_object_scope(CUSTOM_OBJECTS):
        model = ks.models.load_model('model/path')

"""
from graph_attention_student.training import NoLoss
from graph_attention_student.training import ExplanationLoss

CUSTOM_OBJECTS = {
    'NoLoss': NoLoss,
    'ExplanationLoss': ExplanationLoss
}